"""
This module contains the tests to check peleffy's parameters.
"""

import pytest

import tempfile
from simtk import unit
import numpy as np

from .utils import (SET_OF_LIGAND_PATHS, apply_PELE_dihedral_equation,
                    apply_OFF_dihedral_equation, check_CHO_charges,
                    compare_files)
from peleffy.topology import Molecule, Bond, Angle, Proper
from peleffy.template.impact import (WritableBond, WritableAngle,
                                     WritableProper, WritableImproper)
from peleffy.forcefield import OpenForceField, OPLS2005ForceField
from peleffy.forcefield.parameters import BaseParameterWrapper
from peleffy.topology import Topology
from peleffy.template import Impact
from peleffy.utils import get_data_file_path, temporary_cd


FORCEFIELD_NAME = 'openff_unconstrained-1.2.0.offxml'


class TestWrapper(object):
    """
    It contains all the tests that validate the parameter wrapper class.
    """

    def test_forcefield_name_assignment(self):
        """
        It validates the force field name assignment of the parameter
        wrapper.
        """
        from peleffy.forcefield.parameters \
            import (BaseParameterWrapper, OpenForceFieldParameterWrapper,
                    OPLS2005ParameterWrapper, OpenFFOPLS2005ParameterWrapper)

        p = BaseParameterWrapper()

        assert p.forcefield_name == '', \
            'Unexpected force field name found in the parameters wrapper'

        p = BaseParameterWrapper(
            forcefield_name='openff_unconstrained-1.2.1.offxml')

        assert p.forcefield_name == 'openff_unconstrained-1.2.1.offxml', \
            'Unexpected force field name found in the parameters wrapper'

        p = OpenForceFieldParameterWrapper()

        assert p.forcefield_name == 'OpenFF', \
            'Unexpected force field name found in the parameters wrapper'

        p = OpenForceFieldParameterWrapper(
            forcefield_name='openff_unconstrained-1.2.1.offxml')

        assert p.forcefield_name == 'openff_unconstrained-1.2.1.offxml', \
            'Unexpected force field name found in the parameters wrapper'

        p = OPLS2005ParameterWrapper()

        assert p.forcefield_name == 'OPLS2005', \
            'Unexpected force field name found in the parameters wrapper'

        p = OpenForceFieldParameterWrapper(forcefield_name='opls')

        assert p.forcefield_name == 'opls', \
            'Unexpected force field name found in the parameters wrapper'

        p = OpenFFOPLS2005ParameterWrapper()

        assert p.forcefield_name == 'Openff + OPLS2005', \
            'Unexpected force field name found in the parameters wrapper'

        p = OpenForceFieldParameterWrapper(
            forcefield_name='openff_unconstrained-1.2.1.offxml')

        assert p.forcefield_name == 'openff_unconstrained-1.2.1.offxml', \
            'Unexpected force field name found in the parameters wrapper'

    def test_comparison(self):
        """It tests the comparison between parameter wrapper."""
        from peleffy.forcefield.parameters import BaseParameterWrapper

        p = BaseParameterWrapper()
        p2 = BaseParameterWrapper()
        p3 = BaseParameterWrapper(
            forcefield_name='openff_unconstrained-1.2.1.offxml')
        p4 = BaseParameterWrapper({'atom_names': ' C1 '})
        p5 = BaseParameterWrapper({'atom_names': ' C1 '})
        p6 = BaseParameterWrapper(
            {'atom_names': ' C1 '},
            forcefield_name='openff_unconstrained-1.2.1.offxml')

        assert p == p2, 'Unexpected inequality between \'p\' and \'p2\''
        assert p != p3, 'Unexpected equality between \'p\' and \'p3\''
        assert p != p4, 'Unexpected equality between \'p\' and \'p4\''
        assert p4 == p5, 'Unexpected inequality between \'p4\' and \'p5\''
        assert p4 != p6, 'Unexpected equality between \'p4\' and \'p6\''

    def test_to_string(self):
        """It tests the string representation of the parameter wrapper."""

        pdb_path = get_data_file_path('ligands/methane.pdb')
        molecule = Molecule(pdb_path)
        ff = OpenForceField('openff_unconstrained-1.2.1.offxml')
        parameters = ff.parameterize(molecule,
                                     charge_method='gasteiger')
        p_string = parameters.to_string()

        ref_string_file = get_data_file_path(
            'tests/MET_parameters_to_string.txt')
        with open(ref_string_file) as f:
            ref_string = f.read().strip('\n')

        p_lines = p_string.split('\n')
        ref_lines = ref_string.split('\n')

        assert len(p_lines) == len(ref_lines), \
            'Unexpected number of lines: ' \
            + str(len(p_lines)) + ', expected ' + str(len(ref_lines))

        for p_line, ref_line in zip(p_lines, ref_lines):
            assert p_line == ref_line, \
                'Unexpected string representation found in line: ' \
                + '\'{}\' '.format(p_line) \
                + 'which does not match with \'{}\''.format(ref_line)

    def test_to_json(self):
        """It tests the json representation of the parameter wrapper."""

        with tempfile.TemporaryDirectory() as tmpdir:
            with temporary_cd(tmpdir):
                pdb_path = get_data_file_path('ligands/methane.pdb')
                molecule = Molecule(pdb_path)
                ff = OpenForceField('openff_unconstrained-1.2.1.offxml')
                parameters = ff.parameterize(molecule,
                                             charge_method='gasteiger')
                parameters.to_json('parameters_to_check.json')

                ref_json_file = get_data_file_path(
                    'tests/MET_parameters_to_json.json')
                compare_files('parameters_to_check.json', ref_json_file)

    def test_from_json(self):
        """It tests the loading function from json files into the parameter
        wrapper"""

        def compare_BaseParameterWrapper(params1, params2):
            """
            It compares two BaseParameterWrapper objects.
            """
            import numpy as np
            for p in params1.keys():
                if p == 'charges':      #charges have to be handle differently
                    assert params1[p].__eq__(params2[p]).all() == \
                            np.full((len(params1[p])), True, dtype=bool).all()
                else:
                    assert params1[p] == params2[p]

        with tempfile.TemporaryDirectory() as tmpdir:
            with temporary_cd(tmpdir):
                #Force Field to parameterize the molecules
                ff = OpenForceField('openff_unconstrained-1.2.1.offxml')

                # Test for malonate
                m = Molecule(get_data_file_path('ligands/malonate.pdb'))
                params_ref = ff.parameterize(m)
                params_ref.to_json('parameters_mal.json')

                from peleffy.forcefield.parameters import \
                                OpenForceFieldParameterWrapper
                wrapper_off = OpenForceFieldParameterWrapper()
                params_load = wrapper_off.from_json('parameters_mal.json')

                compare_BaseParameterWrapper(params_ref, params_load)

                # Test for methane
                m = Molecule(get_data_file_path('ligands/methane.pdb'))
                params_ref = ff.parameterize(m)
                params_ref.to_json('parameters_met.json')

                from peleffy.forcefield.parameters import \
                                OpenForceFieldParameterWrapper
                wrapper_off = OpenForceFieldParameterWrapper()
                params_load = wrapper_off.from_json('parameters_met.json')

                compare_BaseParameterWrapper(params_ref, params_load)

                # Test for ethylene
                m = Molecule(get_data_file_path('ligands/ethylene.pdb'))
                params_ref = ff.parameterize(m)
                params_ref.to_json('parameters_etl.json')

                from peleffy.forcefield.parameters import \
                                OpenForceFieldParameterWrapper
                wrapper_off = OpenForceFieldParameterWrapper()
                params_load = wrapper_off.from_json('parameters_etl.json')

                compare_BaseParameterWrapper(params_ref, params_load)

    def test_from_impact_template(self):
        """
        It tests the method to generate a parameter wrapper out of an
        impact template.
        """
        def test_generate_OpenForceFieldParameterWrapper(molecule,
                                                         impact_template_path):
            """
            It tests the method to return a OpenForceFieldParameterWrapper
            object from an Impact file by comparing the reference Impact file
            and the generated from the obtained Wrapper.
            """

            from peleffy.forcefield.parameters import \
                OpenForceFieldParameterWrapper

            with tempfile.TemporaryDirectory() as tmpdir:
                with temporary_cd(tmpdir):

                    # Assign parameters to BaseParametersWrapper object from
                    # an Impact Template
                    wrapper_off = OpenForceFieldParameterWrapper()
                    parameters = wrapper_off.from_impact_template(
                        molecule, impact_template_path)

                    # Generate the new impact template
                    topology = Topology(molecule, parameters)
                    impact = Impact(topology)
                    impact.to_file('unlz_generated')

                    # Compare the reference and generated Impact templates
                    compare_files(impact_template_path, 'unlz_generated')

        def test_generate_OPLS2005ParameterWrapper(molecule,
                                                   impact_template_path):
            """
            It tests the method to return a OPLS2005FieldParameterWrapper
            object from an Impact file by comparing the reference Impact file
            and the generated from the obtained Wrapper.
            """

            from peleffy.forcefield.parameters import OPLS2005ParameterWrapper

            with tempfile.TemporaryDirectory() as tmpdir:
                with temporary_cd(tmpdir):

                    # Assign parameters to BaseParametersWrapper object from
                    # an Impact Template
                    wrapper_opls = OPLS2005ParameterWrapper()
                    parameters = wrapper_opls.from_impact_template(
                        molecule, impact_template_path)

                    # Generate the new impact template
                    topology = Topology(molecule, parameters)
                    impact = Impact(topology)
                    impact.to_file('unlz_generated')

                    # Compare the reference and generated Impact templates
                    compare_files(impact_template_path, 'unlz_generated')

        # Test with OFF parametrization (for ethylene)
        pdb_path = get_data_file_path('ligands/ethylene.pdb')
        molecule = Molecule(pdb_path, tag='ETL')
        impact_template_path = get_data_file_path('tests/etlz')
        test_generate_OpenForceFieldParameterWrapper(molecule,
                                                     impact_template_path)

        # Test with OFF parametrization (for methane)
        pdb_path = get_data_file_path('ligands/methane.pdb')
        molecule = Molecule(pdb_path)
        impact_template_path = get_data_file_path('tests/metz')
        test_generate_OpenForceFieldParameterWrapper(molecule,
                                                     impact_template_path)

        # Test with OFF parametrization (for malonate)
        pdb_path = get_data_file_path('ligands/malonate.pdb')
        molecule = Molecule(pdb_path)
        impact_template_path = get_data_file_path('tests/malz')
        test_generate_OpenForceFieldParameterWrapper(molecule,
                                                     impact_template_path)

        # Test with OPLS parametrization (for ethylene)
        pdb_path = get_data_file_path('ligands/ethylene.pdb')
        molecule = Molecule(pdb_path, tag='ETL')
        impact_template_path = get_data_file_path('tests/OPLS_etlz')
        test_generate_OPLS2005ParameterWrapper(molecule,
                                               impact_template_path)

        # Test with molecule with phase different than 0 or 180
        # Load molecule, parameterize and generate Impact template
        molecule = Molecule(smiles='c1c(c(n(n1)S(=O)(=O)C))O')
        impact_template_path = get_data_file_path('tests/unlz')
        test_generate_OpenForceFieldParameterWrapper(molecule,
                                                     impact_template_path)

        # The molecule and Impact template do no represent the same chemical
        # entity
        with pytest.raises(ValueError):
            pdb_path = get_data_file_path('ligands/ethylene.pdb')
            molecule = Molecule(pdb_path, tag='ETL')
            impact_template_path = get_data_file_path('tests/metz')
            test_generate_OpenForceFieldParameterWrapper(molecule,
                                                         impact_template_path)


class TestBonds(object):
    """
    It wraps all tests that involve bond parameters.
    """

    def test_OFF_parameters(self):
        """
        It checks the standard Open Force Field parameterization for
        bonds.
        """

        MAX_THRESHOLD = 1e-3

        # Load benzene ring
        molecule = Molecule(smiles='c1ccccc1')

        # Parameterize
        ff = OpenForceField(FORCEFIELD_NAME)
        parameters = ff.parameterize(molecule, charge_method='gasteiger')

        # Check resulting parameters
        expected_lengths = {(0, 1): 1.389761064076,
                            (0, 5): 1.389761064076,
                            (0, 6): 1.085503378387,
                            (1, 2): 1.389761064076,
                            (1, 7): 1.085503378387,
                            (2, 3): 1.389761064076,
                            (2, 8): 1.085503378387,
                            (3, 4): 1.389761064076,
                            (3, 9): 1.085503378387,
                            (4, 5): 1.389761064076,
                            (4, 10): 1.085503378387,
                            (5, 11): 1.085503378387}

        expected_ks = {(0, 1): 672.0064645264,
                       (0, 5): 672.0064645264,
                       (0, 6): 808.4160937,
                       (1, 2): 672.0064645264,
                       (1, 7): 808.4160937,
                       (2, 3): 672.0064645264,
                       (2, 8): 808.4160937,
                       (3, 4): 672.0064645264,
                       (3, 9): 808.4160937,
                       (4, 5): 672.0064645264,
                       (4, 10): 808.4160937,
                       (5, 11): 808.4160937}

        for bond in parameters['bonds']:
            indexes = (bond['atom1_idx'], bond['atom2_idx'])
            expected_length = unit.Quantity(expected_lengths[indexes],
                                            unit.angstrom)
            expected_k = unit.Quantity(expected_ks[indexes],
                                       unit.kilocalorie
                                       / (unit.angstrom ** 2 * unit.mole))

            assert bond['eq_dist'] - expected_length \
                < unit.Quantity(MAX_THRESHOLD, unit.angstrom), \
                'Invalid length for bond {}'.format(bond)
            assert bond['spring_constant'] - expected_k \
                < unit.Quantity(MAX_THRESHOLD, unit.kilocalorie
                                / (unit.angstrom ** 2 * unit.mole)), \
                'Invalid k for bond {}'.format(bond)

    def test_Impact_writable_parameters(self):
        """
        It checks the Impact writable representation for bonds.
        """

        # Load benzene ring
        molecule = Molecule(smiles='CC=O')

        # Parameterize
        ff = OpenForceField(FORCEFIELD_NAME)
        parameters = ff.parameterize(molecule, charge_method='gasteiger')

        # Generate topology
        topology = Topology(molecule, parameters)

        expected_parameters = list([[1, 2, 332.5750972667, 1.523640340452],
                                    [1, 4, 376.8940758588, 1.094223427522],
                                    [1, 5, 376.8940758588, 1.094223427522],
                                    [1, 6, 376.8940758588, 1.094223427522],
                                    [2, 3, 608.3286693405, 1.225108345696],
                                    [2, 7, 404.20804685, 1.085503378387]])

        # Check resulting parameters
        for bond in topology.bonds:
            w_bond = WritableBond(bond)
            w_parameters = [attr[1] for attr in list(w_bond)]
            assert w_parameters in expected_parameters, \
                'Invalid writable bond parameters {}'.format(w_parameters)


class TestAngles(object):
    """
    It wraps all tests that involve angle parameters.
    """

    def test_OFF_parameters(self):
        """
        It checks the standard Open Force Field parameterization for
        angles.
        """

        MAX_THRESHOLD = 1e-3

        # Load benzene ring
        molecule = Molecule(smiles='c1ccccc1')

        # Parameterize
        ff = OpenForceField(FORCEFIELD_NAME)
        parameters = ff.parameterize(molecule, charge_method='gasteiger')

        # Check resulting parameters
        expected_angles = {(0, 1, 2): 128.2771922378,
                           (0, 1, 7): 133.1339832262,
                           (0, 5, 4): 128.2771922378,
                           (0, 5, 11): 133.1339832262,
                           (1, 0, 5): 128.2771922378,
                           (1, 0, 6): 133.1339832262,
                           (1, 2, 3): 128.2771922378,
                           (1, 2, 8): 133.1339832262,
                           (2, 1, 7): 133.1339832262,
                           (2, 3, 4): 128.2771922378,
                           (2, 3, 9): 133.1339832262,
                           (3, 2, 8): 133.1339832262,
                           (3, 4, 5): 128.2771922378,
                           (3, 4, 10): 133.1339832262,
                           (4, 3, 9): 133.1339832262,
                           (4, 5, 11): 133.1339832262,
                           (5, 0, 6): 133.1339832262,
                           (5, 4, 10): 133.1339832262}

        expected_ks = {(0, 1, 2): 157.3576298529,
                       (0, 1, 7): 68.40592742547,
                       (0, 5, 4): 157.3576298529,
                       (0, 5, 11): 68.40592742547,
                       (1, 0, 5): 157.3576298529,
                       (1, 0, 6): 68.40592742547,
                       (1, 2, 3): 157.3576298529,
                       (1, 2, 8): 68.40592742547,
                       (2, 1, 7): 68.40592742547,
                       (2, 3, 4): 157.3576298529,
                       (2, 3, 9): 68.40592742547,
                       (3, 2, 8): 68.40592742547,
                       (3, 4, 5): 157.3576298529,
                       (3, 4, 10): 68.40592742547,
                       (4, 3, 9): 68.40592742547,
                       (4, 5, 11): 68.40592742547,
                       (5, 0, 6): 68.40592742547,
                       (5, 4, 10): 68.40592742547}

        for angle in parameters['angles']:
            indexes = (angle['atom1_idx'], angle['atom2_idx'],
                       angle['atom3_idx'])
            expected_angle = unit.Quantity(expected_angles[indexes],
                                           unit.degree)
            expected_k = unit.Quantity(expected_ks[indexes],
                                       unit.kilocalorie
                                       / (unit.radian ** 2 * unit.mole))

            assert angle['eq_angle'] - expected_angle \
                < unit.Quantity(MAX_THRESHOLD, unit.degree), \
                'Invalid length for angle {}'.format(angle)
            assert angle['spring_constant'] - expected_k \
                < unit.Quantity(MAX_THRESHOLD, unit.kilocalorie
                                / (unit.radian ** 2 * unit.mole)), \
                'Invalid k for angle {}'.format(angle)

    def test_Impact_writable_parameters(self):
        """
        It checks the Impact writable representation for angles.
        """

        # Load benzene ring
        molecule = Molecule(smiles='CC=O')

        # Parameterize
        ff = OpenForceField(FORCEFIELD_NAME)
        parameters = ff.parameterize(molecule, charge_method='gasteiger')

        # Generate topology
        topology = Topology(molecule, parameters)

        expected_parameters = list(
            [[1, 2, 3, 78.67881492645, 128.2771922378],
             [1, 2, 7, 34.202963712735, 133.1339832262],
             [2, 1, 4, 50.00415699765, 110.81136453],
             [2, 1, 5, 50.00415699765, 110.81136453],
             [2, 1, 6, 50.00415699765, 110.81136453],
             [3, 2, 7, 34.202963712735, 133.1339832262],
             [4, 1, 5, 33.78875634641, 110.2468561538],
             [4, 1, 6, 33.78875634641, 110.2468561538],
             [5, 1, 6, 33.78875634641, 110.2468561538]])

        # Check resulting parameters
        for angle in topology.angles:
            w_angle = WritableAngle(angle)
            w_parameters = [attr[1] for attr in list(w_angle)]
            assert w_parameters in expected_parameters, \
                'Invalid writable angle parameters {}'.format(w_parameters)


class TestDihedrals(object):
    """
    It wraps all tests that involve dihedral parameters.
    """

    def test_OFF_parameters(self):
        """
        It checks the standard Open Force Field parameterization for
        dihedrals.
        """

        # Load molecule
        molecule = Molecule(smiles='C=CC(=O)O')

        # Parameterize
        ff = OpenForceField(FORCEFIELD_NAME)
        parameters = ff.parameterize(molecule, charge_method='gasteiger')

        # Check resulting parameters for proper torsions
        expected_propers = [(0, 1, 2, 3,
                             unit.Quantity(0.603518062312,
                                           unit.kilocalorie / unit.mole),
                             unit.Quantity(180.0, unit.degree),
                             2, 1.0),
                            (0, 1, 2, 3,
                             unit.Quantity(0.5248455212365,
                                           unit.kilocalorie / unit.mole),
                             unit.Quantity(0.0, unit.degree),
                             3, 1.0),
                            (0, 1, 2, 4,
                             unit.Quantity(0.9350453896311,
                                           unit.kilocalorie / unit.mole),
                             unit.Quantity(180.0, unit.degree),
                             2, 1.0),
                            (1, 2, 4, 8,
                             unit.Quantity(2.529110648699,
                                           unit.kilocalorie / unit.mole),
                             unit.Quantity(180.0, unit.degree),
                             2, 1.0),
                            (2, 1, 0, 5,
                             unit.Quantity(5.376019778605,
                                           unit.kilocalorie / unit.mole),
                             unit.Quantity(180.0, unit.degree),
                             2, 1.0),
                            (2, 1, 0, 6,
                             unit.Quantity(5.376019778605,
                                           unit.kilocalorie / unit.mole),
                             unit.Quantity(180.0, unit.degree),
                             2, 1.0),
                            (3, 2, 1, 7,
                             unit.Quantity(0.9350453896311,
                                           unit.kilocalorie / unit.mole),
                             unit.Quantity(180.0, unit.degree),
                             2, 1.0),
                            (3, 2, 4, 8,
                             unit.Quantity(2.237928151469,
                                           unit.kilocalorie / unit.mole),
                             unit.Quantity(180.0, unit.degree),
                             2, 1.0),
                            (3, 2, 4, 8,
                             unit.Quantity(1.23728649144,
                                           unit.kilocalorie / unit.mole),
                             unit.Quantity(0.0, unit.degree),
                             1, 1.0),
                            (4, 2, 1, 7,
                             unit.Quantity(0.9350453896311,
                                           unit.kilocalorie / unit.mole),
                             unit.Quantity(180.0, unit.degree),
                             2, 1.0),
                            (5, 0, 1, 7,
                             unit.Quantity(5.376019778605,
                                           unit.kilocalorie / unit.mole),
                             unit.Quantity(180.0, unit.degree),
                             2, 1.0),
                            (6, 0, 1, 7,
                             unit.Quantity(5.376019778605,
                                           unit.kilocalorie / unit.mole),
                             unit.Quantity(180.0, unit.degree),
                             2, 1.0)]

        assert len(expected_propers) == len(parameters['propers']), \
            'Unexpected number of proper torsions'

        for proper in parameters['propers']:
            proper_parameters = (proper['atom1_idx'], proper['atom2_idx'],
                                 proper['atom3_idx'], proper['atom4_idx'],
                                 proper['k'], proper['phase'],
                                 proper['periodicity'], proper['idivf'])

            assert proper_parameters in expected_propers, \
                'Unexpected proper torsion'

        # Check resulting parameters for improper torsions
        expected_impropers = [(0, 1, 2, 7,
                               unit.Quantity(1.1,
                                             unit.kilocalorie / unit.mole),
                               unit.Quantity(180.0, unit.degree),
                               2, 1),
                              (1, 0, 5, 6,
                               unit.Quantity(1.1,
                                             unit.kilocalorie / unit.mole),
                               unit.Quantity(180.0, unit.degree),
                               2, 1),
                              (1, 2, 3, 4,
                               unit.Quantity(10.5,
                                             unit.kilocalorie / unit.mole),
                               unit.Quantity(180.0, unit.degree),
                               2, 1)]

        assert len(expected_impropers) == \
            len(parameters['impropers']), \
            'Unexpected number of improper torsions'

        for improper in parameters['impropers']:
            improper_parameters = (improper['atom1_idx'],
                                   improper['atom2_idx'],
                                   improper['atom3_idx'],
                                   improper['atom4_idx'],
                                   improper['k'],
                                   improper['phase'],
                                   improper['periodicity'],
                                   improper['idivf'])

            assert improper_parameters in expected_impropers, \
                'Unexpected improper torsion'

    def test_Impact_writable_parameters(self):
        """
        It checks the Impact writable representation for dihedrals.
        """

        # Load benzene ring
        molecule = Molecule(smiles='CC=O')

        # Parameterize
        ff = OpenForceField(FORCEFIELD_NAME)
        parameters = ff.parameterize(molecule, charge_method='gasteiger')

        # Generate topology
        topology = Topology(molecule, parameters)

        expected_parameters = list(
            [[3, 2, 1, 4, 0.5107183999341, 1, 1, 0.0],
             [3, 2, 1, 5, 0.5107183999341, 1, 1, 0.0],
             [3, 2, 1, 6, 0.5107183999341, 1, 1, 0.0],
             [4, 1, 2, 7, 0.1493988458474, 1, 3, 0.0],
             [5, 1, 2, 7, 0.1493988458474, 1, 3, 0.0],
             [6, 1, 2, 7, 0.1493988458474, 1, 3, 0.0],
             [3, 2, 1, 4, -0.0275194074427, 1, 2, 0.0],
             [3, 2, 1, 5, -0.0275194074427, 1, 2, 0.0],
             [3, 2, 1, 6, -0.0275194074427, 1, 2, 0.0],
             [3, 2, 1, 4, -0.1057540923121, -1, 3, 0.0],
             [3, 2, 1, 5, -0.1057540923121, -1, 3, 0.0],
             [3, 2, 1, 6, -0.1057540923121, -1, 3, 0.0]])

        # Check resulting parameters
        for proper in topology.propers:
            w_proper = WritableProper(proper)
            w_parameters = [attr[1] for attr in list(w_proper)]
            assert w_parameters in expected_parameters, \
                'Invalid writable proper parameters {}'.format(w_parameters)

        expected_parameters = list([[1, 2, 3, 7, 1.1, -1, 2]])

        # Check resulting parameters
        for improper in topology.impropers:
            w_improper = WritableImproper(improper)
            w_parameters = [attr[1] for attr in list(w_improper)]
            assert w_parameters in expected_parameters, \
                'Invalid writable improper parameters {}'.format(w_parameters)

    def test_OFF_to_PELE_conversion(self):
        """
        It checks the difference between dihedral equations from PELE and
        Open Force Field. Their values should match throughout all the domain.
        """

        MAX_THRESHOLD = 1e-10

        for ligand_path in SET_OF_LIGAND_PATHS:
            ligand_path = get_data_file_path(ligand_path)

            # Load molecule
            molecule = Molecule(ligand_path)

            # Parameterize
            ff = OpenForceField(FORCEFIELD_NAME)
            parameters = ff.parameterize(molecule, charge_method='gasteiger')

            # Generate topology
            topology = Topology(molecule, parameters)

            x = unit.Quantity(np.arange(0, np.pi, 0.1), unit=unit.radians)

            for PELE_proper, OFF_proper in zip(topology.propers,
                                               topology._OFF_propers):
                PELE_y = apply_PELE_dihedral_equation(PELE_proper, x)
                OFF_y = apply_OFF_dihedral_equation(OFF_proper, x)

                y_diff = PELE_y - OFF_y

                assert np.linalg.norm(y_diff) < MAX_THRESHOLD

    def test_excluded_dihedrals_handler(self):
        """
        It checks the exclusion of dihedrals when generating 1-4 lists
        for PELE. This process is done internally in PELE but we
        need to mark previously those dihedrals to exclude from 1-4
        lists.
        """

        # Load molecule
        molecule = Molecule()

        # Parameterize
        parameters = BaseParameterWrapper()

        # Generate topology
        topology = Topology(molecule, parameters)

        # Add several dummy propers
        topology.add_proper(
            Proper(atom1_idx=0, atom2_idx=1, atom3_idx=2, atom4_idx=3,
                   periodicity=1, prefactor=1, index=0,
                   constant=unit.Quantity(1.0, unit.kilocalorie / unit.mole)))
        topology.add_proper(
            Proper(atom1_idx=0, atom2_idx=1, atom3_idx=2, atom4_idx=3,
                   periodicity=2, prefactor=1, index=1,
                   constant=unit.Quantity(1.0, unit.kilocalorie / unit.mole)))
        topology.add_proper(
            Proper(atom1_idx=0, atom2_idx=1, atom3_idx=5, atom4_idx=3,
                   periodicity=2, prefactor=1, index=2,
                   constant=unit.Quantity(1.0, unit.kilocalorie / unit.mole)))
        topology.add_proper(
            Proper(atom1_idx=0, atom2_idx=1, atom3_idx=2, atom4_idx=4,
                   periodicity=1, prefactor=1, index=3,
                   constant=unit.Quantity(1.0, unit.kilocalorie / unit.mole)))
        topology.add_proper(
            Proper(atom1_idx=0, atom2_idx=1, atom3_idx=2, atom4_idx=5,
                   periodicity=1, prefactor=1, index=4,
                   constant=unit.Quantity(1.0, unit.kilocalorie / unit.mole)))
        topology.add_proper(
            Proper(atom1_idx=0, atom2_idx=1, atom3_idx=2, atom4_idx=6,
                   periodicity=1, prefactor=1, index=5,
                   constant=unit.Quantity(1.0, unit.kilocalorie / unit.mole)))

        # Add one bond
        topology.add_bond(
            Bond(atom1_idx=0, atom2_idx=4,
                 spring_constant=unit.Quantity(1.0, unit.kilocalorie
                                               / (unit.angstrom ** 2
                                                  * unit.mole)),
                 eq_dist=unit.Quantity(1.0, unit.angstrom)))

        # Add one angle
        topology.add_angle(
            Angle(atom1_idx=0, atom2_idx=1, atom3_idx=5,
                  spring_constant=unit.Quantity(1.0, unit.kilocalorie
                                                / (unit.radian ** 2
                                                   * unit.mole)),
                  eq_angle=unit.Quantity(1.0, unit.degrees)))

        topology._handle_excluded_propers()

        # Assertions
        for proper in topology.propers:
            if proper.index == 0:
                assert proper.exclude is False, \
                    'Proper 0 must be INCLUDED in 1-4 list'
            elif proper.index == 1:
                assert proper.exclude is False, \
                    'Proper 1 must be INCLUDED in 1-4 list'
            elif proper.index == 2:
                assert proper.exclude is True, \
                    'Proper 2 must be EXCLUDED in 1-4 list'
            elif proper.index == 3:
                assert proper.exclude is True, \
                    'Proper 3 must be EXCLUDED in 1-4 list'
            elif proper.index == 4:
                assert proper.exclude is True, \
                    'Proper 4 must be EXCLUDED in 1-4 list'
            elif proper.index == 5:
                assert proper.exclude is False, \
                    'Proper 5 must be INCLUDED in 1-4 list'

    def test_nonstandard_dihedrals_writable_parameters(self):
        """
        It checks the writable representation of non standard dihedrals.
        """
        # Load molecule
        molecule = Molecule(smiles='c1c(c(n(n1)S(=O)(=O)C))O')

        # Parameterize
        ff = OpenForceField(FORCEFIELD_NAME)
        parameters = ff.parameterize(molecule, charge_method='gasteiger')

        # Generate topology
        topology = Topology(molecule, parameters)

        expected_parameters = [[1, 2, -3, 4, 5.376019778605, -1, 2, 0.0],
                               [1, 2, 3, 12, 5.376019778605, -1, 2, 0.0],
                               [1, 2, 10, 16, 0.8722932201352, -1, 2, 0.0],
                               [1, 5, -4, 3, -0.1847566338874, 1, 3, 0.0],
                               [1, 5, 4, 6, -0.1847566338874, 1, 3, 0.0],
                               [2, 1, -5, 4, 6.867249875207, -1, 2, 0.0],
                               [2, 3, -4, 5, 1.131350927723, -1, 2, 0.0],
                               [2, 3, 4, 6, 1.131350927723, -1, 2, 0.0],
                               [3, 2, -1, 5, 0.9350453896311, -1, 2, 0.0],
                               [3, 2, 1, 11, 0.9350453896311, -1, 2, 0.0],
                               [3, 2, 10, 16, 0.8722932201352, -1, 2, 0.0],
                               [3, 4, 6, 7, 0.5146229154488, 1, 1, 0.0],
                               [3, 4, 6, 8, 0.5146229154488, 1, 1, 0.0],
                               [3, 4, 6, 9, -0.5050335923881, 1, 3, 90.0],
                               [4, 3, 2, 10, 5.376019778605, -1, 2, 0.0],
                               [4, 5, 1, 11, 6.867249875207, -1, 2, 0.0],
                               [4, 6, 9, 13, 0.1055132590818, 1, 3, 0.0],
                               [4, 6, 9, 14, 0.1055132590818, 1, 3, 0.0],
                               [4, 6, 9, 15, 0.1055132590818, 1, 3, 0.0],
                               [5, 1, 2, 10, 0.9350453896311, -1, 2, 0.0],
                               [5, 4, 3, 12, 1.131350927723, -1, 2, 0.0],
                               [5, 4, 6, 7, 0.082636589537, 1, 1, 0.0],
                               [5, 4, 6, 8, 0.082636589537, 1, 1, 0.0],
                               [5, 4, 6, 9, 1.914473924133, 1, 1, 0.0],
                               [6, 4, 3, 12, 1.131350927723, -1, 2, 0.0],
                               [7, 6, 9, 13, 0.1055132590818, 1, 3, 0.0],
                               [7, 6, 9, 14, 0.1055132590818, 1, 3, 0.0],
                               [7, 6, 9, 15, 0.1055132590818, 1, 3, 0.0],
                               [8, 6, 9, 13, 0.1055132590818, 1, 3, 0.0],
                               [8, 6, 9, 14, 0.1055132590818, 1, 3, 0.0],
                               [8, 6, 9, 15, 0.1055132590818, 1, 3, 0.0],
                               [10, 2, 1, 11, 0.9350453896311, -1, 2, 0.0],
                               [10, 2, 3, 12, 5.376019778605, -1, 2, 0.0],
                               [1, 5, -4, 3, 1.471695489046, -1, 2, 0.0],
                               [1, 5, 4, 6, 1.471695489046, -1, 2, 0.0],
                               [3, 4, 6, 9, 0.3649469393198, 1, 2, 0.0]]

        # Check resulting parameters
        for proper in topology.propers:
            w_proper = WritableProper(proper)
            w_parameters = [attr[1] for attr in list(w_proper)]
            assert w_parameters in expected_parameters, \
                'Invalid writable proper parameters {}'.format(w_parameters)

        expected_parameters = [[1, 2, 3, 10, 1.1, -1, 2],
                               [2, 1, 5, 11, 1.1, -1, 2],
                               [2, 3, 4, 12, 1.1, -1, 2],
                               [3, 4, 5, 6, 10.5, -1, 2]]

        # Check resulting parameters
        for improper in topology.impropers:
            w_improper = WritableImproper(improper)
            w_parameters = [attr[1] for attr in list(w_improper)]
            assert w_parameters in expected_parameters, \
                'Invalid writable improper parameters {}'.format(w_parameters)

    def test_OPLS_dummy_propers(self):
        """
        It checks that the number of propers that are obtained from OPLS
        is correct, ensuring that dummy propers (propers with a null
        force constant whose definitions are only used in the 1-4 lists
        of PELE) are present in the parameterized proper list.
        """
        from peleffy.forcefield.parameters import OPLS2005ParameterWrapper

        # Load molecule
        molecule = Molecule(
            get_data_file_path('ligands/octafluorocyclobutane.pdb'))

        # Set force field and obtain parameters
        ffld_file = get_data_file_path('tests/CO1_ffld_output.txt')
        with open(ffld_file) as f:
            ffld_output = f.read()

        parameters = OPLS2005ParameterWrapper.from_ffld_output(molecule,
                                                               ffld_output)

        # Generate topology
        topology = Topology(molecule, parameters)

        # Validate number of propers
        assert len(topology.propers) == 100, \
            'Unexpected number of proper torsions'


class TestCharges(object):
    """
    It wraps all tests that involve charge parameters.
    """
    LIGAND_PATH = 'ligands/oleic_acid.pdb'

    def test_am1bcc_method(self):
        """It tests the am1bcc method"""
        ligand_path = get_data_file_path(self.LIGAND_PATH)

        # Load molecule
        molecule = Molecule(ligand_path)

        # Parameterize
        ff = OpenForceField(FORCEFIELD_NAME)
        parameters = ff.parameterize(molecule, charge_method='am1bcc')

        # Check charges
        check_CHO_charges(parameters)

    def test_gasteiger_method(self):
        """It tests the gasteiger method"""
        ligand_path = get_data_file_path(self.LIGAND_PATH)

        # Load molecule
        molecule = Molecule(ligand_path)

        # Parameterize
        ff = OpenForceField(FORCEFIELD_NAME)
        parameters = ff.parameterize(molecule, charge_method='gasteiger')

        # Check charges
        check_CHO_charges(parameters)

    def test_OPLS_method(self):
        """It tests the OPLS method"""

        ligand_path = get_data_file_path(self.LIGAND_PATH)

        # Load molecule
        molecule = Molecule(ligand_path)

        expected_charges = [-0.22, 0.7, -0.12, -0.8, -0.8, -0.12, -0.12,
                            -0.12, -0.12, -0.12, -0.115, -0.115, -0.12,
                            -0.12, -0.12, -0.12, -0.12, -0.12, -0.12,
                            -0.18, 0.06, 0.06, 0.06, 0.06, 0.06, 0.06,
                            0.06, 0.06, 0.06, 0.06, 0.06, 0.06, 0.06,
                            0.06, 0.115, 0.115, 0.06, 0.06, 0.06, 0.06,
                            0.06, 0.06, 0.06, 0.06, 0.06, 0.06, 0.06,
                            0.06, 0.06, 0.06, 0.06, 0.06, 0.06]

        # Workaround to avoid the use of the Schrodinger Toolkit
        from .utils import parameterize_opls2005

        # Load OPLS2005 force field and locate ffld_file
        oplsff = OPLS2005ForceField()
        ffld_file = get_data_file_path('tests/OLC_ffld_output.txt')

        # Parameterize
        parameters = parameterize_opls2005(oplsff, molecule, ffld_file)

        partial_charges = parameters['charges']

        assert len(partial_charges) == len(expected_charges), \
            'Size of Molecule\'s partial charges is expected to match ' \
            + 'with size of reference charges list'

        for partial_charges, expected_charge in zip(partial_charges,
                                                    expected_charges):
            assert partial_charges == unit.Quantity(expected_charge,
                                                    unit.elementary_charge), \
                'Unexpected partial charge {}'.format(partial_charges)

    def test_dummy_method(self):
        """It tests the dummy charge calculator."""

        ligand_path = get_data_file_path(self.LIGAND_PATH)

        # Load molecule
        molecule = Molecule(ligand_path)

        expected_charges = [0.0, ] * 53

        # Parameterize
        ff = OpenForceField(FORCEFIELD_NAME)
        parameters = ff.parameterize(molecule, charge_method='dummy')

        partial_charges = parameters['charges']

        assert len(partial_charges) == len(expected_charges), \
            'Size of Molecule\'s partial charges is expected to match ' \
            + 'with size of reference charges list'

        for partial_charges, expected_charge in zip(partial_charges,
                                                    expected_charges):
            assert partial_charges == unit.Quantity(expected_charge,
                                                    unit.elementary_charge), \
                'Unexpected partial charge {}'.format(partial_charges)


class TestSolventParameters(object):
    """
    It holds tests to validate the assignment of solvent parameters.
    """

    def test_add_SGBNP_solvent_parameters(self):
        """
        It tests the function that adds the SGBNP solvent parameters to
        the OPLSParameters collection.
        """

        from simtk import unit
        from peleffy.forcefield.parameters import OPLS2005ParameterWrapper

        # Using a standard atom type
        params1 = OPLS2005ParameterWrapper(
            {'atom_names': [' C1 ', ' H1 ', ' H2 ', ' H3 ', ' H4 '],
             'atom_types': ['CT', 'HC', 'HC', 'HC', 'HC'],
             'charges': [-0.24, 0.06, 0.06, 0.06, 0.06],
             'sigmas': [3.5, 2.5, 2.5, 2.5, 2.5],
             'epsilons': [0.066, 0.03, 0.03, 0.03, 0.03]})

        # Using a similar atom type
        params2 = OPLS2005ParameterWrapper(
            {'atom_names': [' C1 ', ' H1 ', ' H2 ', ' H3 ', ' H4 '],
             'atom_types': ['C3M', 'HC', 'HC', 'HC', 'HC'],
             'charges': [-0.24, 0.06, 0.06, 0.06, 0.06],
             'sigmas': [3.5, 2.5, 2.5, 2.5, 2.5],
             'epsilons': [0.066, 0.03, 0.03, 0.03, 0.03]})

        # Using a default atom type
        params3 = OPLS2005ParameterWrapper(
            {'atom_names': [' C1 ', ' H1 ', ' H2 ', ' H3 ', ' H4 '],
             'atom_types': ['XX', 'HC', 'HC', 'HC', 'HC'],
             'charges': [-0.24, 0.06, 0.06, 0.06, 0.06],
             'sigmas': [3.5, 2.5, 2.5, 2.5, 2.5],
             'epsilons': [0.066, 0.03, 0.03, 0.03, 0.03]})

        OPLS2005ParameterWrapper._add_SGBNP_solvent_parameters(params1)
        OPLS2005ParameterWrapper._add_SGBNP_solvent_parameters(params2)
        OPLS2005ParameterWrapper._add_SGBNP_solvent_parameters(params3)

        assert params1['SGB_radii'][0] == \
            unit.Quantity(1.975, unit.angstrom), 'Unexpected SGB radius'
        assert params1['vdW_radii'][0] == \
            unit.Quantity(1.750, unit.angstrom), 'Unexpected vdW radius'
        assert params1['gammas'][0] == 0.005000000, 'Unexpected gamma'
        assert params1['alphas'][0] == -0.741685710, 'Unexpected alpha'

        assert params2['SGB_radii'][0] == \
            unit.Quantity(2.002, unit.angstrom), 'Unexpected SGB radius'
        assert params2['vdW_radii'][0] == \
            unit.Quantity(1.775, unit.angstrom), 'Unexpected vdW radius'
        assert params2['gammas'][0] == 0.023028004, 'Unexpected gamma'
        assert params2['alphas'][0] == -0.852763146, 'Unexpected alpha'

        assert params3['SGB_radii'][0] == \
            unit.Quantity(1.500, unit.angstrom), 'Unexpected SGB radius'
        assert params3['vdW_radii'][0] == \
            unit.Quantity(1.250, unit.angstrom), 'Unexpected vdW radius'
        assert params3['gammas'][0] == 0.005000000, 'Unexpected gamma'
        assert params3['alphas'][0] == 0.000000000, 'Unexpected alpha'

    def test_GBSA_params_by_type(self):
        """
        It tests the params_by_type dictionary in the assignment of the
        GBSA parameters.
        """

        from peleffy.forcefield.parameters import OPLS2005ParameterWrapper

        # 1st test
        OPLS_params = OPLS2005ParameterWrapper()

        # Create mock molecule containing just one customized atom
        OPLS_params['atom_names'] = [' C1 ']
        OPLS_params['atom_types'] = [' CT']
        degree_by_name = dict(((' C1 ', 4), ))
        parent_by_name = dict(((' C1 ', None), ))
        element_by_name = dict(((' C1 ', 'C'), ))

        OPLS2005ParameterWrapper._add_GBSA_solvent_parameters(OPLS_params,
                                                              degree_by_name,
                                                              parent_by_name,
                                                              element_by_name)

        assert OPLS_params['GBSA_radii'] == \
            [unit.Quantity(0.72, unit.angstrom)], \
            'Unexpected GBSA radii'
        assert OPLS_params['GBSA_scales'] == [1.9], \
            'Unexpected GBSA scale'

        # 2nd test
        OPLS_params = OPLS2005ParameterWrapper()

        # Create mock molecule containing just one customized atom
        OPLS_params['atom_names'] = [' C1 ']
        OPLS_params['atom_types'] = [' CW']
        degree_by_name = dict(((' C1 ', 3), ))
        parent_by_name = dict(((' C1 ', None), ))
        element_by_name = dict(((' C1 ', 'C'), ))

        OPLS2005ParameterWrapper._add_GBSA_solvent_parameters(OPLS_params,
                                                              degree_by_name,
                                                              parent_by_name,
                                                              element_by_name)

        assert OPLS_params['GBSA_radii'] == \
            [unit.Quantity(0.72, unit.angstrom)], \
            'Unexpected GBSA radii'
        assert OPLS_params['GBSA_scales'] == [1.875], \
            'Unexpected GBSA scale'

        # 3rd test
        OPLS_params = OPLS2005ParameterWrapper()

        # Create mock molecule containing just one customized atom
        OPLS_params['atom_names'] = [' O1 ']
        OPLS_params['atom_types'] = [' O ']
        degree_by_name = dict(((' O1 ', 1), ))
        parent_by_name = dict(((' O1 ', None), ))
        element_by_name = dict(((' O1 ', 'O'), ))

        OPLS2005ParameterWrapper._add_GBSA_solvent_parameters(OPLS_params,
                                                              degree_by_name,
                                                              parent_by_name,
                                                              element_by_name)

        assert OPLS_params['GBSA_radii'] == \
            [unit.Quantity(0.85, unit.angstrom)], \
            'Unexpected GBSA radii'
        assert OPLS_params['GBSA_scales'] == [1.48], \
            'Unexpected GBSA scale'

        # 4th test
        OPLS_params = OPLS2005ParameterWrapper()

        # Create mock molecule containing just one customized atom
        OPLS_params['atom_names'] = [' H10']
        OPLS_params['atom_types'] = [' H2']
        degree_by_name = dict(((' H10', 1), ))
        parent_by_name = dict(((' H10', 'C'), ))
        element_by_name = dict(((' H10', 'H'), ))

        OPLS2005ParameterWrapper._add_GBSA_solvent_parameters(OPLS_params,
                                                              degree_by_name,
                                                              parent_by_name,
                                                              element_by_name)

        assert OPLS_params['GBSA_radii'] == \
            [unit.Quantity(0.85, unit.angstrom)], \
            'Unexpected GBSA radii'
        assert OPLS_params['GBSA_scales'] == [1.25], \
            'Unexpected GBSA scale'

    def test_GBSA_params_by_element(self):
        """
        It tests the scale_by_element and radius_by_element dictionaries
        in the assignment of the GBSA parameters.
        """

        from peleffy.forcefield.parameters import OPLS2005ParameterWrapper

        # 1st test
        OPLS_params = OPLS2005ParameterWrapper()

        # Create mock molecule containing just one customized atom
        OPLS_params['atom_names'] = [' C1 ']
        OPLS_params['atom_types'] = [' C?!']
        degree_by_name = dict(((' C1 ', 3), ))
        parent_by_name = dict(((' C1 ', None), ))
        element_by_name = dict(((' C1 ', 'C'), ))

        OPLS2005ParameterWrapper._add_GBSA_solvent_parameters(OPLS_params,
                                                              degree_by_name,
                                                              parent_by_name,
                                                              element_by_name)

        assert OPLS_params['GBSA_radii'] == \
            [unit.Quantity(0.72, unit.angstrom)], \
            'Unexpected GBSA radii'
        assert OPLS_params['GBSA_scales'] == [1.875], \
            'Unexpected GBSA scale'

        # 2nd test
        OPLS_params = OPLS2005ParameterWrapper()

        # Create mock molecule containing just one customized atom
        OPLS_params['atom_names'] = [' O1 ']
        OPLS_params['atom_types'] = ['O?!']
        degree_by_name = dict(((' O1 ', 2), ))
        parent_by_name = dict(((' O1 ', None), ))
        element_by_name = dict(((' O1 ', 'O'), ))

        OPLS2005ParameterWrapper._add_GBSA_solvent_parameters(OPLS_params,
                                                              degree_by_name,
                                                              parent_by_name,
                                                              element_by_name)

        assert OPLS_params['GBSA_radii'] == \
            [unit.Quantity(0.85, unit.angstrom)], \
            'Unexpected GBSA radii'
        assert OPLS_params['GBSA_scales'] == [1.535], \
            'Unexpected GBSA scale'

        # 3rd test
        OPLS_params = OPLS2005ParameterWrapper()

        # Create mock molecule containing just one customized atom
        OPLS_params['atom_names'] = [' O1 ']
        OPLS_params['atom_types'] = ['O?!']
        degree_by_name = dict(((' O1 ', 1), ))
        parent_by_name = dict(((' O1 ', None), ))
        element_by_name = dict(((' O1 ', 'O'), ))

        OPLS2005ParameterWrapper._add_GBSA_solvent_parameters(OPLS_params,
                                                              degree_by_name,
                                                              parent_by_name,
                                                              element_by_name)

        assert OPLS_params['GBSA_radii'] == \
            [unit.Quantity(0.85, unit.angstrom)], \
            'Unexpected GBSA radii'
        assert OPLS_params['GBSA_scales'] == [1.48], \
            'Unexpected GBSA scale'

        # 4th test
        OPLS_params = OPLS2005ParameterWrapper()

        # Create mock molecule containing just one customized atom
        OPLS_params['atom_names'] = [' H10']
        OPLS_params['atom_types'] = ['H?!']
        degree_by_name = dict(((' H10', 1), ))
        parent_by_name = dict(((' H10', 'C'), ))
        element_by_name = dict(((' H10', 'H'), ))

        OPLS2005ParameterWrapper._add_GBSA_solvent_parameters(OPLS_params,
                                                              degree_by_name,
                                                              parent_by_name,
                                                              element_by_name)

        assert OPLS_params['GBSA_radii'] == \
            [unit.Quantity(0.85, unit.angstrom)], \
            'Unexpected GBSA radii'
        assert OPLS_params['GBSA_scales'] == [1.25], \
            'Unexpected GBSA scale'

        # 5th test
        OPLS_params = OPLS2005ParameterWrapper()

        # Create mock molecule containing just one customized atom
        OPLS_params['atom_names'] = [' H10']
        OPLS_params['atom_types'] = ['H?!']
        degree_by_name = dict(((' H10', 1), ))
        parent_by_name = dict(((' H10', 'N'), ))
        element_by_name = dict(((' H10', 'H'), ))

        OPLS2005ParameterWrapper._add_GBSA_solvent_parameters(OPLS_params,
                                                              degree_by_name,
                                                              parent_by_name,
                                                              element_by_name)

        assert OPLS_params['GBSA_radii'] == \
            [unit.Quantity(0.85, unit.angstrom)], \
            'Unexpected GBSA radii'
        assert OPLS_params['GBSA_scales'] == [1.15], \
            'Unexpected GBSA scale'

        # 6th test
        OPLS_params = OPLS2005ParameterWrapper()

        # Create mock molecule containing just one customized atom
        OPLS_params['atom_names'] = [' H10']
        OPLS_params['atom_types'] = ['H?!']
        degree_by_name = dict(((' H10', 1), ))
        parent_by_name = dict(((' H10', 'O'), ))
        element_by_name = dict(((' H10', 'H'), ))

        OPLS2005ParameterWrapper._add_GBSA_solvent_parameters(OPLS_params,
                                                              degree_by_name,
                                                              parent_by_name,
                                                              element_by_name)

        assert OPLS_params['GBSA_radii'] == \
            [unit.Quantity(0.85, unit.angstrom)], \
            'Unexpected GBSA radii'
        assert OPLS_params['GBSA_scales'] == [1.05], \
            'Unexpected GBSA scale'

        # 7th test
        OPLS_params = OPLS2005ParameterWrapper()

        # Create mock molecule containing just one customized atom
        OPLS_params['atom_names'] = [' C1 ']
        OPLS_params['atom_types'] = [' C?!']
        degree_by_name = dict(((' C1 ', 2), ))
        parent_by_name = dict(((' C1 ', None), ))
        element_by_name = dict(((' C1 ', 'C'), ))

        OPLS2005ParameterWrapper._add_GBSA_solvent_parameters(OPLS_params,
                                                              degree_by_name,
                                                              parent_by_name,
                                                              element_by_name)

        assert OPLS_params['GBSA_radii'] == \
            [unit.Quantity(0.72, unit.angstrom)], \
            'Unexpected GBSA radii'
        assert OPLS_params['GBSA_scales'] == [1.825], \
            'Unexpected GBSA scale'

    def test_GBSA_default_params(self):
        """
        It tests the default parameters of the GBSA implementation.
        """
        import io
        from peleffy.forcefield.parameters import OPLS2005ParameterWrapper
        from peleffy.utils import Logger

        OPLS_params = OPLS2005ParameterWrapper()

        # Create mock molecule containing just one customized atom
        OPLS_params['atom_names'] = [' C1 ']
        OPLS_params['atom_types'] = [' C?!']
        degree_by_name = dict(((' C1 ', 2), ))
        parent_by_name = dict(((' C1 ', None), ))
        element_by_name = dict(((' C1 ', '?'), ))

        import logging

        # Force a hard reset of logging library and the logger it manages
        from importlib import reload
        logging.shutdown()
        reload(logging)

        # Initiate logger
        log = Logger()

        # Try the default level (INFO)
        # Catch logger messages to string buffer
        with io.StringIO() as buf:
            # Add custom handler to logger
            log_handler = logging.StreamHandler(buf)
            log._logger.handlers = list()
            log._logger.addHandler(log_handler)

            OPLS2005ParameterWrapper._add_GBSA_solvent_parameters(
                OPLS_params, degree_by_name, parent_by_name, element_by_name)

            # Get string from buffer
            output = buf.getvalue()

            assert output == 'Warning: OBC parameters for ' \
                + 'C1 C?! NOT found in the template database. ' \
                + 'Using default parameters\n'

        assert OPLS_params['GBSA_radii'] == \
            [unit.Quantity(0.80, unit.angstrom)], \
            'Unexpected GBSA radii'
        assert OPLS_params['GBSA_scales'] == [2.0], \
            'Unexpected GBSA scale'

    def test_add_GBSA_solvent_parameters(self):
        """
        It tests the function that adds the GBSA solvent parameters to
        the OPLSParameters collection.
        """

        from peleffy.forcefield.parameters import OPLS2005ParameterWrapper
        from peleffy.topology import Molecule
        from peleffy.utils import get_data_file_path

        def generate_and_compare_GBSA_solvent_parameters(pdb_file,
                                                         ffld_file,
                                                         reference_txt):
            """
            Given a ligand, it tests that the HTC radii and the scale factor
            for heteroatoms computed with the function
            _add_GBSA_solvent_parameters correspond to the ones
            obtained with the original solventOBCParamsGenerator.py script.

            Parameters
            ----------
            pdb_file : str
                The path to the PDB of the ligand to test
            ffld_file : str
                The path to the ffld_server's output file
            reference_txt : str
                The path to reference TXT file obtained
                from solventOBCParamsGenerator.py
            """

            # Load the molecule
            molecule = Molecule(pdb_file)

            # Set force field and obtain parameters
            with open(ffld_file) as f:
                ffld_output = f.read()

            # Initializate wrapper for OPLS2005 parameters
            wrapper_opls = OPLS2005ParameterWrapper()
            parameters = wrapper_opls.from_ffld_output(molecule, ffld_output)

            # Get the radi and scale parameters for each atom name
            atom_names = [param.replace('_', '')
                          for param in parameters['atom_names']]
            d_radii = dict(zip(atom_names, parameters['GBSA_radii']))
            d_scale = dict(zip(atom_names, parameters['GBSA_scales']))

            # Load the reference file obtained from solventOBCParamsGenerator.py
            data = open(reference_txt, 'r').readlines()

            # Check that the parameters correspond
            for line in data:
                params = line.split()
                assert float(params[3]) == d_scale.get(params[1]), \
                    'Unexpected GBSA Overlap factor'
                assert unit.Quantity(float(params[4]), unit.angstrom) == \
                    d_radii.get(params[1]), \
                    'Unexpected GBSA radius'

        # For malonate
        ffld_file = get_data_file_path('tests/MAL_ffld_output.txt')
        pdb_file = get_data_file_path('ligands/malonate.pdb')
        reference_txt = get_data_file_path('tests/malz_OBCParams.txt')
        generate_and_compare_GBSA_solvent_parameters(pdb_file,
                                                     ffld_file,
                                                     reference_txt)

        # For methane
        ffld_file = get_data_file_path('tests/MET_ffld_output.txt')
        pdb_file = get_data_file_path('ligands/methane.pdb')
        reference_txt = get_data_file_path('tests/metz_OBCParams.txt')
        generate_and_compare_GBSA_solvent_parameters(pdb_file,
                                                     ffld_file,
                                                     reference_txt)

        # For ethylene
        ffld_file = get_data_file_path('tests/ETL_ffld_output.txt')
        pdb_file = get_data_file_path('ligands/ethylene.pdb')
        reference_txt = get_data_file_path('tests/etlz_OBCParams.txt')
        generate_and_compare_GBSA_solvent_parameters(pdb_file,
                                                     ffld_file,
                                                     reference_txt)



"""
This module provides classes to define and construct force field
parameters.
"""


__all__ = ["OpenForceFieldParameterWrapper",
           "OPLS2005ParameterWrapper",
           "OpenFFOPLS2005ParameterWrapper"]


from collections import defaultdict
from simtk import unit

from peleffy.utils import get_data_file_path
from peleffy.utils import Logger


class BaseParameterWrapper(dict):
    """
    It defines the base class of a parameter wrapper that inherits from
    dict.
    """

    _name = ''
    _fixed_keys = ('atom_names', 'atom_types', 'charges', 'sigmas',
                   'epsilons', 'SGB_radii', 'vdW_radii', 'gammas',
                   'alphas', 'GBSA_radii', 'GBSA_scales')
    _unfixed_keys = ('bonds', 'angles', 'propers', 'impropers')
    _attribute_restrictions = {'bonds': ('atom1_idx', 'atom2_idx',
                                         'spring_constant', 'eq_dist'),
                               'angles': ('atom1_idx', 'atom2_idx',
                                          'atom3_idx', 'spring_constant',
                                          'eq_angle'),
                               'propers': ('atom1_idx', 'atom2_idx',
                                           'atom3_idx', 'atom4_idx',
                                           'periodicity', 'phase',
                                           'k', 'idivf'),
                               'impropers': ('atom1_idx', 'atom2_idx',
                                             'atom3_idx', 'atom4_idx',
                                             'periodicity', 'phase',
                                             'k', 'idivf')}
    _keys = _fixed_keys + _unfixed_keys

    def __init__(self, parameters_dict=dict(), forcefield_name=None):
        """
        It initializes a parameter wrapper object.

        Parameters
        ----------
        parameters_dict : dict
             A dictionary keyed by parameter type that contains the
             parameters to store. Default is an empty dict
        forcefield_name : str
            The name of the force field the parameters belong to
        """
        for key in self._keys:
            self[key] = list()

        for key, value in parameters_dict.items():
            self[key] = value

        if forcefield_name is None:
            self._forcefield_name = self._name
        else:
            self._forcefield_name = forcefield_name

    def __eq__(self, other):
        """
        It sets the equality operator for the BaseParameterWrapper class.

        Parameters
        ----------
        other : a BaseParameterWrapper object
            The other BaseParameterWrapper object to compare with the
            current one
        """
        if not isinstance(other, BaseParameterWrapper):
            return False

        return super().__eq__(other) and \
            self.forcefield_name == other.forcefield_name

    def __ne__(self, other):
        """
        It sets the inequality operator for the BaseParameterWrapper class.

        Parameters
        ----------
        other : a BaseParameterWrapper object
            The other BaseParameterWrapper object to compare with the
            current one
        """
        return not self.__eq__(other)

    def __setitem__(self, key, val):
        """
        It sets an item in the dictionary, only if the key is an
        expected key and fulfills the size and attribute restrictions.

        Parameters
        ----------
        key : str
            The parameter type
        val : list[float, int or simtk.unit.Quantity]
            The list of corresponding values
        """
        if key not in self._keys:
            raise KeyError('An unexpected parameter type cannot be set '
                           + 'as a dictionary key. Unexpected type: '
                           + '{}'.format(key))

        okay = True
        if key in self._fixed_keys:
            for local_key in self.keys():
                if local_key not in self._fixed_keys:
                    continue
                local_size = len(self[local_key])
                val_size = len(val)
                if local_size != 0 and val_size != 0:
                    if len(val) != local_size:
                        okay = False

        if not okay:
            raise ValueError('Supplied values for key \'{}\' '.format(key)
                             + 'are not of the same size as values that '
                             + 'have already been loaded. Input length: '
                             + '{} Expected length: '.format(len(val))
                             + '{}'.format(local_size))

        okay = True
        if key in self._attribute_restrictions:
            attributes = self._attribute_restrictions[key]
            if isinstance(val, dict):
                if len(attributes) != len(val.keys()):
                    okay = False
                for attribute in attributes:
                    if attribute not in val.keys():
                        okay = False
            elif isinstance(val, list):
                for val_element in val:
                    if len(attributes) != len(val_element.keys()):
                        okay = False
                    for attribute in attributes:
                        if attribute not in val_element.keys():
                            okay = False
            else:
                raise ValueError('Unexpected input value with key '
                                 + '\'{}\' '.format(key)
                                 + 'and value \'{}\''.format(val))
        if not okay:
            raise ValueError('Supplied values for key \'{}\' '.format(key)
                             + 'do not contain the expected attributes. '
                             + 'They are: '
                             + '\'{}\''.format(', '.join(attributes)))

        dict.__setitem__(self, key, val)

    def __str__(self):
        """
        It returns the string representation of this parameter wrapper.

        Returns
        -------
        string_representation : str
            The string representation
        """
        return self.to_string()

    def add_parameters(self, label, parameters):
        """
        It adds a list of parameters of the same type to the collection.

        Parameters
        ----------
        label : str
            The label to describe the parameter type
        parameters : list
            The list of parameters to include to the collection
        """
        self[label] = parameters

    def to_string(self):
        """
        It returns a string representation of the parameters stored in
        the wrapper.

        Returns
        -------
        string_repr : str
            The string representation
        """

        from peleffy.utils import convert_all_quantities_to_string
        import pprint

        string_repr = pprint.pformat(convert_all_quantities_to_string(self))

        return string_repr

    def to_json(self, output_path):
        """
        It saves this parameter wrapper as a json file.

        Parameters
        ----------
        output_path : str
            The path to save the output json file
        """

        import json
        from peleffy.utils import convert_all_quantities_to_string

        with open(output_path, 'w') as f:
            json.dump(convert_all_quantities_to_string(self), f,
                      indent=4, sort_keys=True)

    def from_json(self, input_path):
        """
        It loads a json file containing the parameters into the parameter
        wrapper.

        Parameters
        ----------
        input_path : str
            The path to the json file to load the parameters from

        Returns
        -------
        self : a BaseParameterWrapper object
            The resulting parameters wrapper
        """
        import json
        def correct_type(label, value):
            """
            It converts the parameters loaded from the JSON file into the
            expected data type in the parameter wrapper.

            Parameters
            ----------
            label : str
                Label for the parameters
            value : list
                Set of parameters

            Returns
            -------
            correct_value : list
                Set of paramaeters with the expected data type
            """
            import simtk.unit
            from peleffy.utils import string_to_quantity

            # Dictionary relating parameters key with the expected data type
            dict_units = {
                'alphas': float, 'gammas': float,
                'charges':  simtk.unit.quantity.Quantity,
                'sigmas' : 'list_Quantity', 'epsilons' : 'list_Quantity',
                'SGB_radii' : 'list_Quantity','vdW_radii' : 'list_Quantity',
                'angles': dict, 'bonds': dict, 'impropers': dict,
                'propers': dict, 'GBSA_radii': 'list_Quantity',
                'GBSA_scales': 'list_Quantity',
                'atom_names': str, 'atom_types' : str
                }

            # Skip data type transformation if the list is empty or None values
            if value == [None,] * len(value):
                return value
            if value == []:
                return None

            # Perform data type transformation, if needed
            else:
                if dict_units[label] is str:
                    correct_value = value

                if dict_units[label] is float:
                    correct_value  = [float(v) if not v is None else v
                    for v in value]

                if dict_units[label] is simtk.unit.quantity.Quantity:
                    correct_value = string_to_quantity(value)
                    return correct_value

                if dict_units[label] == 'list_Quantity':
                    correct_value = \
                            [unit.Quantity(value=string_to_quantity(v)._value,
                                           unit=string_to_quantity(v).unit)
                            for v in value]

                if dict_units[label] is dict:
                    correct_value = []
                    for element in value:
                        correct_dict = dict()
                        for k,v in element.items():
                            if 'idx' in k:
                                correct_dict[k] = int(v)
                            else:
                                try:
                                    correct_dict[k] = unit.Quantity(
                                            value=string_to_quantity(v)._value,
                                            unit=string_to_quantity(v).unit)
                                except:
                                    correct_dict[k] = v
                        correct_value.append(correct_dict)
                return correct_value

        # Load the dict from the JSON file
        with open(input_path) as f:
            data = json.load(f)

        # Correct the data type format and fetch the BaseParameterWrapper
        for key, value in data.items():
            value_correct = correct_type(key, value)
            if not value_correct is None:
                self.add_parameters(key, value_correct)
        return self

    @property
    def atom_iterator(self):
        """
        It returns an iterator that goes through all atom parameters.
        The order of parameters that is followed is: 'atom_names',
        'atom_types', 'sigmas', 'epsilons', 'charges', 'SGB_radii',
        'vdW_radii', 'gammas', 'alphas'.

        Returns
        -------
        iterator : a zip object
            It is an iterator that returns all atom parameters following
            their order
        """
        assert len(self['atom_names']) == len(self['atom_types']) \
            and len(self['atom_names']) == len(self['sigmas']) \
            and len(self['atom_names']) == len(self['epsilons']) \
            and len(self['atom_names']) == len(self['charges']) \
            and len(self['atom_names']) == len(self['SGB_radii']) \
            and len(self['atom_names']) == len(self['vdW_radii']) \
            and len(self['atom_names']) == len(self['gammas']) \
            and len(self['atom_names']) == len(self['alphas']), \
            'Size of atom parameter lists should match'

        return zip(self['atom_names'], self['atom_types'],
                   self['sigmas'], self['epsilons'],
                   self['charges'], self['SGB_radii'],
                   self['vdW_radii'], self['gammas'],
                   self['alphas'])

    def is_empty(self):
        """
        If the are no parameters stored in this wrapper, it returns True.

        Returns
        -------
        answer : bool
            Whether this parameter wrapper is empty or not
        """
        return all(len(self[key]) == 0 for key in self)

    @property
    def name(self):
        """
        It returns the name of the current parameter wrapper.

        Returns
        -------
        name : str
            The name of the parameter wrapper
        """
        return self._name

    @staticmethod
    def from_impact_template(molecule, impact_template_path):
        """
        It returns a parameter wrapper out of an impact template.

        Parameters
        ----------
        molecule : a peleffy.topology.Molecule
            The peleffy's Molecule object
        impact_template_path : str
            The path to the impact template from where the parameters
            will be fetched

        Returns
        -------
        params : a BaseParameterWrapper object
            The resulting parameters wrapper

        Examples
        --------

        Obtain a parameter wrapper from an Impact template.

        >>> from peleffy.topology import Molecule

        >>> molecule = Molecule('molecule.pdb')
        >>> impact_template_path = 'molz'

        >>> from peleffy.forcefield.parameters import \
                OpenForceFieldParametersWrapper

        >>> wrapper_off = OpenForceFieldParametersWrapper()
        >>> parameters = wrapper_off.from_impact_template(
                molecule, impact_template_path)

        """

        def index(atom_idx):
            """
            Atom's index.

            Parameters
            ----------
            index : int
                The Atom's index for a WritableWrapper

            Returns
            -------
            index : int
                The Atom's index for a BaseParameterWrapper object
            """
            return abs(int(atom_idx)) - 1

        def zero_to_none(values):
            """
            It converts a vector of zeros to None.

            Parameters
            ----------
            values : list
                List of parameters.

            Returns
            -------
            values : list
                It is set to None if all the parameters are zeros.
            """
            n_atoms = len(values)
            if type(values[0]) == float:
                if all(value == float(0.0) for value in values):
                    values = [None, ] * n_atoms
            if type(values[0]) == unit.quantity.Quantity:
                if all(value == unit.Quantity(0, unit.angstroms) for value
                       in values):
                    values = values = [None, ] * n_atoms
            return values

        def get_phase(info):
            """
            It computes the phase of a Dihedral given its prefactor. PELE
            prefactors can be  equal to 1 or -1. If the phase constant is
            different than 0 or 180, an extra column in the Impact template
            specifies its value.

            Parameters
            ----------
            info : list
                List of parameters for this Dihedral

            Returns
            -------
            phase : simtk.unit.Quantity
                The phase constant of the Dihedral
            """
            if len(info) > 7:
                return unit.Quantity(
                    value=float(info[7]),
                    unit=unit.degree)
            else:
                prefactor = float(info[5])
                if prefactor == 1:
                    return unit.Quantity(
                        value=0.00,
                        unit=unit.degree)
                if prefactor == -1:
                    return unit.Quantity(
                        value=180.00,
                        unit=unit.degree)

        def reindex_atom_idx(list_params, dict_index):
            """
            It sorts and reindexes atoms of a parameters list according to a
            dictionary.

            Parameters
            ----------
            list_params : list
                List of parameters to reindex
            dict_index : dict
                Dictionary that defines the reindexer between old and new index

            Returns
            -------
            list_params : list
                Updated list of parameters with the correct indexes
            """
            for i, atom_info in enumerate(list_params):
                for key, value in atom_info.items():
                    if '_idx' in key:
                        list_params[i][key] = dict_index.get(value)
            return list_params

        # Parse Impact template file
        tag, nbon, bond, thet, phi, iphi = ([] for i in range(6))
        with open(impact_template_path, 'r') as fd:
            type_info = 'tag'
            for line in fd.readlines():
                if not line.startswith('*'):
                    if 'NBON' in line:
                        type_info = 'nbon'
                    elif 'BOND' in line:
                        type_info = 'bond'
                    elif 'THET' in line:
                        type_info = 'thet'
                    elif 'PHI' in line and 'IPHI' not in line:
                        type_info = 'phi'
                    elif 'IPHI' in line:
                        type_info = 'iphi'
                    else:
                        if type_info == 'tag':
                            tag.append(line)
                        if type_info == 'nbon':
                            nbon.append(line)
                        if type_info == 'bond':
                            bond.append(line)
                        if type_info == 'thet':
                            thet.append(line)
                        if type_info == 'phi':
                            phi.append(line)
                        if type_info == 'iphi':
                            iphi.append(line)

        # Atom names and atom types
        atom_names_list, atom_types_list = ([] for i in range(2))
        for line in tag[1:]:
            info = line.split()
            atom_names_list.append(info[4])
            atom_types_list.append(info[3])

        # Charges, epsilons, sigmas, SGB radius, vdW radius, gammas and alphas
        charges_list, epsilons_list, sigmas_list, SGB_list, vdW_list, \
            gammas_list, alphas_list = ([] for i in range(7))
        for line in nbon:
            info = line.split()
            charges_list.append(unit.Quantity(
                value=float(info[3]),
                unit=unit.elementary_charge))
            epsilons_list.append(unit.Quantity(
                value=float(info[2]),
                unit=unit.kilocalorie / unit.mole))
            sigmas_list.append(unit.Quantity(
                value=float(info[1]),
                unit=unit.angstrom))
            SGB_list.append(unit.Quantity(
                value=float(info[4]),
                unit=unit.angstrom))
            vdW_list.append(unit.Quantity(
                value=float(info[5]),
                unit=unit.angstrom))
            gammas_list.append(float(info[6]))
            alphas_list.append(float(info[7]))

        # Bonds
        bonds_list = []
        for line in bond:
            info = line.split()
            case = {'atom1_idx': index(info[0]),
                    'atom2_idx': index(info[1]),
                    'eq_dist': unit.Quantity(
                        value=float(info[3]),
                        unit=unit.angstrom),
                    'spring_constant': unit.Quantity(
                        value=float(info[2]),
                        unit=unit.kilocalorie
                        / (unit.angstrom ** 2 * unit.mole))}
            bonds_list.append(case)

        # Angles
        angles_list = []
        for line in thet:
            info = line.split()
            case = {'atom1_idx': index(info[0]),
                    'atom2_idx': index(info[1]),
                    'atom3_idx': index(info[2]),
                    'eq_angle': unit.Quantity(
                        value=float(info[4]),
                        unit=unit.degrees),
                    'spring_constant': unit.Quantity(
                        value=float(info[3]),
                        unit=unit.kilocalorie
                        / (unit.radian ** 2 * unit.mole))}
            angles_list.append(case)

        # Impropers
        impropers_list = []
        for line in iphi[:-1]:
            info = line.split()
            case = {'atom1_idx': index(info[0]),
                    'atom2_idx': index(info[1]),
                    'atom3_idx': index(info[2]),
                    'atom4_idx': index(info[3]),
                    'idivf': abs(float(info[5])),
                    'k': unit.Quantity(
                        value=float(info[4]),
                        unit=unit.kilocalorie / unit.mole),
                    'periodicity': int(float(info[6])),
                    'phase': get_phase(info)}
            impropers_list.append(case)

        # Propers
        propers_list = []
        for line in phi:
            info = line.split()
            case = {'atom1_idx': index(info[0]),
                    'atom2_idx': index(info[1]),
                    'atom3_idx': index(info[2]),
                    'atom4_idx': index(info[3]),
                    'idivf': abs(float(info[5])),
                    'k': unit.Quantity(
                        value=float(info[4]),
                        unit=unit.kilocalorie / unit.mole),
                    'periodicity': int(float(info[6])),
                    'phase': get_phase(info)}
            propers_list.append(case)

        # Initialize the BaseParameterWrapper object
        params = BaseParameterWrapper()

        # PDB atom names from Molecule object
        ordered_pdb_atom_names = [pdb_name.replace(' ', '_') for pdb_name in
                                  molecule.get_pdb_atom_names()]

        # Check up that the Molecule representation and the Impact template
        # represent the same chemical entity
        if not set(ordered_pdb_atom_names) == set(atom_names_list):
            raise ValueError(
                "The Impact template file {} ".format(impact_template_path) +
                "does not represent the same chemical entity as the molecule.")

        # Molecule object and Impact template have the atoms in the same order
        if ordered_pdb_atom_names == atom_names_list:
            # Assign parameters from Impact template to the BaseParameterWrapper
            # object
            params['angles'] = angles_list
            params['atom_names'] = atom_names_list
            params['atom_types'] = atom_types_list
            params['bonds'] = bonds_list
            params['charges'] = charges_list
            params['epsilons'] = epsilons_list
            params['impropers'] = impropers_list
            params['propers'] = propers_list
            params['sigmas'] = sigmas_list
            params['vdW_radii'] = vdW_list
            params['SGB_radii'] = zero_to_none(SGB_list)
            params['gammas'] = zero_to_none(gammas_list)
            params['alphas'] = zero_to_none(alphas_list)

        # Molecule object and Impact template have the atoms in different order
        if ordered_pdb_atom_names != atom_names_list:
            # Reindexer dictionary to sort atoms as Molecule object
            index_order = [atom_names_list.index(atom_name) for atom_name in
                           ordered_pdb_atom_names]
            dict_reindex = dict(zip(index_order, range(len(index_order))))

            # Assign parameters from Impact template to the BaseParameterWrapper
            # object reordering atoms according to Molecule object atoms
            params['atom_names'] = [atom_names_list[i] for i in index_order]
            params['atom_types'] = [atom_types_list[i] for i in index_order]
            params['charges'] = [charges_list[i] for i in index_order]
            params['epsilons'] = [epsilons_list[i] for i in index_order]
            params['sigmas'] = [sigmas_list[i] for i in index_order]
            params['vdW_radii'] = [vdW_list[i] for i in index_order]
            params['SGB_radii'] = [zero_to_none(SGB_list)[i] for i in
                                   index_order]
            params['gammas'] = [zero_to_none(gammas_list)[i] for i in
                                index_order]
            params['alphas'] = [zero_to_none(alphas_list)[i] for i in
                                index_order]
            params['impropers'] = reindex_atom_idx(impropers_list,
                                                   dict_reindex)
            params['propers'] = reindex_atom_idx(propers_list,
                                                 dict_reindex)
            params['angles'] = reindex_atom_idx(angles_list,
                                                dict_reindex)
            params['bonds'] = reindex_atom_idx(bonds_list,
                                               dict_reindex)
        return params

    @property
    def forcefield_name(self):
        return self._forcefield_name


class OpenForceFieldParameterWrapper(BaseParameterWrapper):
    """
    It defines a parameters wrapper for an OpenFF force field.
    """

    _name = 'OpenFF'

    @staticmethod
    def from_label_molecules(molecule, parameters, openff_version):
        """
        It parses the parameters coming from the label_molecules()
        method from OpenFF toolkit.

        Parameters
        ----------
        molecule : an peleffy.topology.Molecule
            The peleffy's Molecule object
        parameters_dict : dict
             A dictionary keyed by parameter type that contains the
             parameters to store
        openff_version : str
            The version of the OpenFF force field the parameters belong to

        Returns
        -------
        parameters : an OpenForceFieldParameterWrapper object
            The resulting parameters wrapper
        """

        I_6R_OF_2 = 0.8908987181403393  # The inverse of the sixth root of 2

        def build_dict(parameters, attribute_name):
            """
            It builds the dictionary of parameters of a specific force type.

            Parameters
            ----------
            parameters : dict[tuple, openforcefield.typing.engines.smirnoff.parameters.ParameterHandler]
                The parameters of a specific force type grouped by tuples
                with the atom ids that the parameters belong to
            attribute_name : str
                The name of the attribute that is requested

            Returns
            -------
            value_by_index : dict[tuple, parameter_value]
                The parameter values that were requested grouped by the atom
                ids they belong to (arranged as a tuple)
            """
            if parameters:
                value_by_index = dict()
                for index, parameter in parameters.items():
                    value_by_index[index] = getattr(parameter, attribute_name)

                return value_by_index

        def build_dynamic_dicts(parameters, attr_core_name):
            """
            It builds the dynamically the dictionaries of parameters of a
            specific force type.

            It works with the same idea as _build_dict(), however it can
            handle multiple dictionary definitions were consecutive
            parameters of the same type are found in the force type's
            parameters dictionary. It works, for example, with the multiple
            proper and improper definitions found in the OpenForceField API.
            More information in the <ProperTorsions> and <ImproperTorsions>
            sections at:
            https://open-forcefield-toolkit.readthedocs.io/en/0.7.0/smirnoff.html

            Parameters
            ----------
            parameters : dict[tuple, openforcefield.typing.engines.smirnoff.parameters.ParameterHandler]
                The parameters of a specific force type grouped by tuples
                with the atom ids that the parameters belong to
            attribute_name : str
                The name of the attribute that is requested

            Returns
            -------
            value_by_index : dict[tuple, parameter_value]
                The parameter values that were requested grouped by the atom
                ids they belong to (arranged as a tuple)
            """
            if parameters:
                parameters_by_index = defaultdict(dict)
                all_attr_ids_found = set()
                for index, parameter in parameters.items():
                    counter = int(1)
                    attr_name = attr_core_name + str(counter)
                    while(attr_name in parameter.to_dict()):
                        all_attr_ids_found.add(counter)
                        attr_value = getattr(parameter, attr_name)
                        parameters_by_index[index][counter] = attr_value
                        counter += 1
                        attr_name = attr_core_name + str(counter)

                output_list = list()
                for attr_id in sorted(all_attr_ids_found):
                    value_by_index = dict()
                    for index in parameters.keys():
                        if attr_id in parameters_by_index[index]:
                            value_by_index[index] = \
                                parameters_by_index[index][attr_id]
                        else:
                            value_by_index[index] = None

                    output_list.append(value_by_index)

                return output_list

        def sigma_from_rmin_half(rmin_half):
            """
            It converts rmin_half values to sigmas according to:
            http://ambermd.org/Questions/vdwequation.pdf
            """
            sigma = I_6R_OF_2 * 2 * rmin_half

            return sigma

        # This will be the dictionary that will contain the parsed data
        params = defaultdict(list)

        pdb_atom_names = molecule.get_pdb_atom_names()
        n_atoms = len(pdb_atom_names)

        # PELE needs underscores instead of whitespaces
        params['atom_names'] = [name.replace(' ', '_') for name
                                in pdb_atom_names]
        params['atom_types'] = ['OFFT', ] * n_atoms

        # TODO: Safety check of the parameters, if any atom has no parameters
        # it should complain here

        # van der Waals parameters
        if 'vdW' in parameters:
            vdW_parameters = parameters['vdW']
            sigma_by_idx = build_dict(vdW_parameters, 'sigma')
            for idx in range(0, n_atoms):
                params['sigmas'].append(sigma_by_idx[(idx, )])

            if all([sigma is None for sigma in params['sigmas']]):
                params['sigmas'] = list()
                rmin_half_by_idx = build_dict(vdW_parameters, 'rmin_half')
                for idx in range(0, n_atoms):
                    params['sigmas'].append(sigma_from_rmin_half(
                        rmin_half_by_idx[(idx, )]))

            epsilon_by_idx = build_dict(vdW_parameters, 'epsilon')
            for idx in range(0, n_atoms):
                params['epsilons'].append(epsilon_by_idx[(idx, )])

            params['SGB_radii'] = [None, ] * n_atoms
            params['vdW_radii'] = [sigma / 2.0 for sigma in params['sigmas']]
            params['gammas'] = [None, ] * n_atoms
            params['alphas'] = [None, ] * n_atoms

        # bond parameters
        if 'Bonds' in parameters:
            bond_parameters = parameters['Bonds']
            length_by_idx = build_dict(bond_parameters, 'length')
            k_by_idx = build_dict(bond_parameters, 'k')
            for idxs in length_by_idx.keys():
                idx1, idx2 = idxs
                params['bonds'].append(
                    {'atom1_idx': idx1,
                     'atom2_idx': idx2,
                     # PELE works with half of the OFF's spring
                     'spring_constant': k_by_idx[idxs] / 2.0,
                     'eq_dist': length_by_idx[idxs]
                     })

        # angle parameters
        if 'Angles' in parameters:
            angle_parameters = parameters['Angles']
            angle_by_idx = build_dict(angle_parameters, 'angle')
            k_by_idx = build_dict(angle_parameters, 'k')
            for idxs in angle_by_idx.keys():
                idx1, idx2, idx3 = idxs
                params['angles'].append(
                    {'atom1_idx': idx1,
                     'atom2_idx': idx2,
                     'atom3_idx': idx3,
                     # PELE works with half of the OFF's spring
                     'spring_constant': k_by_idx[idxs] / 2.0,
                     'eq_angle': angle_by_idx[idxs]
                     })

        # proper parameters
        if 'ProperTorsions' in parameters:
            proper_parameters = parameters['ProperTorsions']
            periodicities = build_dynamic_dicts(proper_parameters,
                                                'periodicity')
            phases = build_dynamic_dicts(proper_parameters, 'phase')
            ks = build_dynamic_dicts(proper_parameters, 'k')
            idivfs = build_dynamic_dicts(proper_parameters, 'idivf')
            if (periodicities is not None and phases is not None
                    and ks is not None and idivfs is not None):
                for periodicity_by_idx, phase_by_idx, k_by_idx, \
                        idivf_by_idx in zip(periodicities, phases, ks,
                                            idivfs):

                    assert periodicity_by_idx.keys() == phase_by_idx.keys() \
                        and periodicity_by_idx.keys() == k_by_idx.keys() \
                        and periodicity_by_idx.keys() == idivf_by_idx.keys(), \
                        'Unconsistent torsional parameter indexes. '\
                        + 'Keys should match.'

                    for idxs in periodicity_by_idx.keys():
                        idx1, idx2, idx3, idx4 = idxs

                        periodicity = periodicity_by_idx[idxs]
                        phase = phase_by_idx[idxs]
                        k = k_by_idx[idxs]
                        idivf = idivf_by_idx[idxs]

                        if (periodicity is not None and phase is not None
                                and k is not None and idivf is not None):
                            params['propers'].append(
                                {'atom1_idx': idx1,
                                 'atom2_idx': idx2,
                                 'atom3_idx': idx3,
                                 'atom4_idx': idx4,
                                 'periodicity': periodicity,
                                 'phase': phase,
                                 'k': k,
                                 'idivf': idivf
                                 })

        # improper parameters
        if 'ImproperTorsions' in parameters:
            improper_parameters = parameters['ImproperTorsions']
            periodicities = build_dynamic_dicts(improper_parameters,
                                                'periodicity')
            phases = build_dynamic_dicts(improper_parameters, 'phase')
            ks = build_dynamic_dicts(improper_parameters, 'k')
            idivfs = build_dynamic_dicts(improper_parameters, 'idivf')
            if (periodicities is not None and phases is not None
                    and ks is not None):
                # idivf is a optional parameter in OpenForceField
                if len(idivfs) == 0:
                    for period_by_index in periodicities:
                        idivfs.append(dict(zip(
                            period_by_index.keys(),
                            [1, ] * len(period_by_index.keys()))))

                for periodicity_by_idx, phase_by_idx, k_by_idx, \
                        idivf_by_idx in zip(periodicities, phases, ks,
                                            idivfs):

                    assert periodicity_by_idx.keys() == phase_by_idx.keys() \
                        and periodicity_by_idx.keys() == k_by_idx.keys() \
                        and periodicity_by_idx.keys() == idivf_by_idx.keys(), \
                        'Unconsistent torsional parameter indexes. '\
                        + 'Keys should match.'

                    for idxs in periodicity_by_idx.keys():
                        idx1, idx2, idx3, idx4 = idxs

                        periodicity = periodicity_by_idx[idxs]
                        phase = phase_by_idx[idxs]
                        k = k_by_idx[idxs]
                        idivf = idivf_by_idx[idxs]

                        if (periodicity is not None and phase is not None
                                and k is not None and idivf is not None):
                            params['impropers'].append(
                                {'atom1_idx': idx1,
                                 'atom2_idx': idx2,
                                 'atom3_idx': idx3,
                                 'atom4_idx': idx4,
                                 'periodicity': periodicity,
                                 'phase': phase,
                                 'k': k,
                                 'idivf': idivf
                                 })

        # GBSA parameters
        if 'GBSA' in parameters:
            gbsa_parameters = parameters['GBSA']
            params['GBSA_radii'] = build_dict(gbsa_parameters, 'radius')
            params['GBSA_scales'] = build_dict(gbsa_parameters, 'scale')

        return OpenForceFieldParameterWrapper(params, openff_version)


class OPLS2005ParameterWrapper(BaseParameterWrapper):
    """
    It defines a parameters wrapper for OPLS2005 force field.
    """

    _name = 'OPLS2005'

    @staticmethod
    def from_ffld_output(molecule, ffld_output):
        """
        It parses the parameters coming from the Schrodinger's
        ffld_server output file.

        Parameters
        ----------
        molecule : an peleffy.topology.Molecule
            The peleffy's Molecule object
        ffld_output : str
            The path to the ffld_server's output file

        Returns
        -------
        parameters : an OpenForceFieldParameterWrapper object
            The resulting parameters wrapper
        """

        from simtk import unit
        from peleffy.utils.toolkits import RDKitToolkitWrapper

        params = defaultdict(list)

        # Assign atom names according to the input PDB file (if any)
        pdb_atom_names = molecule.get_pdb_atom_names()

        # PELE needs underscores instead of whitespaces
        params['atom_names'] = [name.replace(' ', '_') for name
                                in pdb_atom_names]

        section = 'out'
        name_to_index = dict()  # To pair atom names and indexes

        lines_iterator = iter(ffld_output.split('\n'))

        for line in lines_iterator:
            if line.startswith('OPLSAA FORCE FIELD TYPE ASSIGNED'):
                section = 'atoms'

                # Skip the next 3 lines
                next(lines_iterator)
                next(lines_iterator)
                next(lines_iterator)

            elif line.startswith(' Stretch'):
                section = 'bonds'

            elif line.startswith(' Bending'):
                section = 'angles'

            elif line.startswith(' proper Torsion'):
                section = 'propers'

            elif line.startswith(' improper Torsion'):
                section = 'impropers'

            elif line == '':
                continue

            elif section == 'atoms':
                if line.startswith('-'):
                    continue

                fields = line.split()
                assert len(fields) > 7, 'Unexpected number of fields ' \
                    + 'found at line {}'.format(line)

                name_to_index[line[0:8]] = len(name_to_index)

                params['atom_types'].append(fields[3])
                params['charges'].append(
                    unit.Quantity(float(fields[4]),
                                  unit.elementary_charge))
                params['sigmas'].append(
                    unit.Quantity(float(fields[5]),
                                  unit.angstrom))
                params['epsilons'].append(
                    unit.Quantity(float(fields[6]),
                                  unit.kilocalorie / unit.mole))

            elif section == 'bonds':
                fields = line.split()
                assert len(fields) > 4, 'Unexpected number of fields ' \
                    + 'found at line {}'.format(line)

                params['bonds'].append(
                    {'atom1_idx': name_to_index[line[0:8]],
                     'atom2_idx': name_to_index[line[8:16]],
                     'spring_constant': unit.Quantity(
                        float(fields[2]), unit.kilocalorie
                        / (unit.angstrom ** 2 * unit.mole)),
                     'eq_dist': unit.Quantity(float(fields[3]),
                                              unit.angstrom)
                     })

            elif section == 'angles':
                fields = line.split()
                assert len(fields) > 5, 'Unexpected number of fields ' \
                    + 'found at line {}'.format(line)

                params['angles'].append(
                    {'atom1_idx': name_to_index[line[0:8]],
                     'atom2_idx': name_to_index[line[8:16]],
                     'atom3_idx': name_to_index[line[16:24]],
                     'spring_constant': unit.Quantity(
                        float(fields[3]), unit.kilocalorie
                        / (unit.radian ** 2 * unit.mole)),
                     'eq_angle': unit.Quantity(float(fields[4]),
                                               unit.degrees)
                     })

            elif section == 'propers':
                fields = line.split()
                assert len(fields) > 9, 'Unexpected number of fields ' \
                    + 'found at line {}'.format(line)

                atom1_idx = name_to_index[line[0:8]]
                atom2_idx = name_to_index[line[8:16]]
                atom3_idx = name_to_index[line[16:24]]
                atom4_idx = name_to_index[line[24:32]]

                for k, periodicity, phase in zip(
                        fields[4:8], [1, 2, 3, 4],
                        [unit.Quantity(0.0, unit.degree),
                         unit.Quantity(180.0, unit.degree),
                         unit.Quantity(0.0, unit.degree),
                         unit.Quantity(180.0, unit.degree)]):
                    k = float(k)
                    if k != 0.0:
                        k = unit.Quantity(k, unit.kilocalorie / unit.mole)
                        params['propers'].append(
                            {'atom1_idx': atom1_idx,
                             'atom2_idx': atom2_idx,
                             'atom3_idx': atom3_idx,
                             'atom4_idx': atom4_idx,
                             'periodicity': periodicity,
                             'phase': phase,
                             'k': k / 2.0,  # PELE works with half of Schrodinger's force constant
                             'idivf': 1.0
                             })

                # In case all four k's are zero, we still need to include
                # the proper torsion to be used by PELE in the 1-4
                # interactions
                if all([float(k) == 0.0 for k in fields[4:8]]):
                    params['propers'].append(
                        {'atom1_idx': atom1_idx,
                         'atom2_idx': atom2_idx,
                         'atom3_idx': atom3_idx,
                         'atom4_idx': atom4_idx,
                         'periodicity': 1.0,
                         'phase': unit.Quantity(0.0, unit.degree),
                         'k': unit.Quantity(0.0,
                                            unit.kilocalorie / unit.mole),
                         'idivf': 1.0
                         })

            elif section == 'impropers':
                fields = line.split()
                assert len(fields) > 5, 'Unexpected number of fields ' \
                    + 'found at line {}'.format(line)

                k = unit.Quantity(float(fields[4]),
                                  unit.kilocalorie / unit.mole)

                params['impropers'].append(
                    {'atom1_idx': name_to_index[line[0:8]],
                     'atom2_idx': name_to_index[line[8:16]],
                     'atom3_idx': name_to_index[line[16:24]],
                     'atom4_idx': name_to_index[line[24:32]],
                     'periodicity': 2,
                     'phase': unit.Quantity(180.0, unit.degree),
                     'k': k / 2.0,  # PELE works with half of Schrodinger's force constant
                     'idivf': 1.0
                     })

        opls_parameters_wrapper = OPLS2005ParameterWrapper(params)
        OPLS2005ParameterWrapper._add_SGBNP_solvent_parameters(
            opls_parameters_wrapper)

        # Employ RDKit to extract atom degree and parent type lists
        wrapper = RDKitToolkitWrapper()
        atom_names = wrapper.get_atom_names(molecule)
        degree_by_name = dict(zip(atom_names,
                                  wrapper.get_atom_degrees(molecule)))
        parent_by_name = dict(zip(atom_names,
                                  wrapper.get_hydrogen_parents(molecule)))
        element_by_name = dict(zip(atom_names,
                                   wrapper.get_elements(molecule)))

        OPLS2005ParameterWrapper._add_GBSA_solvent_parameters(
            opls_parameters_wrapper, degree_by_name,
            parent_by_name, element_by_name)

        return opls_parameters_wrapper

    @staticmethod
    def _find_similar_atom_types(atom_type, tried):
        """
        It tries to find a similar atom type, skipping the ones that
        have already been tried. It uses the definitions from the
        similarity.param file.

        Parameters
        ----------
        atom_type : str
            The atom type from which similar atom types will be searched
        tried : list[str]
            The list of atom types that have already been tried and
            will be skipped

        Returns
        -------
        new_atom_type : str
            The most similar atom type that has been found, if any.
            Otherwise, it returns None
        """

        new_atom_type = None
        best_similarity = 0
        similarity_path = get_data_file_path(
            'parameters/similarity.param')

        with open(similarity_path) as f:
            for line in f:
                fields = line.split()
                assert len(fields) > 2, 'Unexpected number of fields ' \
                    + 'at line {}'.format(line)

                atom_type1, atom_type2, similarity = fields[0:3]
                if (atom_type == atom_type1
                        and float(similarity) > best_similarity
                        and atom_type2 not in tried):
                    best_similarity = float(similarity)
                    new_atom_type = atom_type2
                elif (atom_type == atom_type2
                        and float(similarity) > best_similarity
                        and atom_type1 not in tried):
                    best_similarity = float(similarity)
                    new_atom_type = atom_type1

        return new_atom_type

    @staticmethod
    def _add_SGBNP_solvent_parameters(OPLS_params):
        """
        It adds the SGBNP solvent parameters (used in the SGBNP solvent
        implemented in the OPLS2005 of PELE) to the OPLS parameters
        collection.

        Parameters
        ----------
        OPLS_params : an OPLS2005ParameterWrapper object
            The set of lists of parameters grouped by parameter type.
            Thus, the dictionary has the following keys: atom_names,
            atom_types, charges, sigmas, and epsilons. The following
            solvent parameters will be added to the collection: SGB_radii,
            vdW_radii, gammas, alphas
        """
        from simtk import unit

        solvent_data = dict()
        parameters_path = get_data_file_path(
            'parameters/f14_sgbnp.param')

        with open(parameters_path) as f:
            for line in f:
                if line.startswith('#'):
                    continue

                fields = line.split()
                assert len(fields) > 7, 'Unexpected line with less ' \
                    'than 8 fields at {}'.format(line)

                atom_type = fields[1]

                solvent_data[atom_type] = {
                    'SGB_radii': unit.Quantity(float(fields[4]),
                                               unit.angstrom),
                    'vdW_radii': unit.Quantity(float(fields[5]),
                                               unit.angstrom),
                    'gammas': float(fields[6]),
                    'alphas': float(fields[7])}

        parameters_to_add = defaultdict(list)
        tried = list()

        for atom_type in OPLS_params['atom_types']:
            parameters_found = False
            while(not parameters_found):
                if atom_type in solvent_data:
                    for label, value in solvent_data[atom_type].items():
                        parameters_to_add[label].append(value)
                    parameters_found = True

                else:
                    new_atom_type = \
                        OPLS2005ParameterWrapper._find_similar_atom_types(
                            atom_type, tried)
                    if new_atom_type is None:
                        atom_type = 'DF'  # Set it to default
                    else:
                        tried.append(new_atom_type)
                        atom_type = new_atom_type

        for label, params in parameters_to_add.items():
            OPLS_params.add_parameters(label, params)

    @staticmethod
    def _add_GBSA_solvent_parameters(OPLS_params, degree_by_name,
                                     parent_by_name,
                                     element_by_name):
        """
        It adds the GBSA solvent parameters (used in the OBC solvent
        implemented in the OPLS2005 of PELE) to the OPLS parameters
        collection.

        Parameters
        ----------
        OPLS_params : an OPLS2005ParameterWrapper object
            The set of lists of parameters grouped by parameter type.
            Thus, the dictionary has the following keys: atom_names,
            atom_types, charges, sigmas, and epsilons. The following
            solvent parameters will be added to the collection: SGB_radii,
            vdW_radii, gammas, alphas
        degree_by_name : dict
            The dictionary containing the number of bonds for
            each atom_name
        parent_by_name : dict
            The dictionary containing the element of the parent
            for hydrogen atoms, keyed by atom name of the child
        element_by_name : dict
            The dictionary containing the atom elements keyed by
            atom name
        """
        import re
        import json

        def _check_bonds(scale, atom_type, degree, parent):
            """
            It checks the number of bonds and the parent atom for
            terminal H. Besides, in some especified cases it updates
            the scale factor.

            Parameters
            ----------
            scale : str
                The scale factor of the current atom
            atom_type : str
                The atom type of the current atom
            degree : int
                The number of bonds of the current atom
            parent : str
                The element belonging to the parent atom, in case that the
                current atom is a hydrogen atom
            """

            if atom_type == 'H' and parent == 'O':
                scale = float(1.05)
            if atom_type == 'H' and parent == 'N':
                scale = float(1.15)
            if atom_type == 'C' and degree == 3:
                scale = float(1.875)
            if atom_type == 'C' and degree == 2:
                scale = float(1.825)
            if atom_type == 'N' and degree == 4:
                scale = float(1.625)
            if atom_type == 'N' and degree == 1:
                scale = float(1.60)
            if atom_type == 'O' and degree == 1:
                scale = float(1.48)

            return scale

        def _find_GBSA_parameters_according_to(atom_name, atom_type,
                                               degree, element, parent):
            """
            It computes the HTC radii and the Overlap factor for Heteroatoms.
            The parameters have been extracted from Tinker Molecular package. If
            one parameter has not been defined in the OBC templates it puts the
            default parameter and raises a warning.

            Parameters
            ----------
            atom_name : str
                Atom name
            atom_type : str
                Atom type
            degree : str
                Number of bonds
            element : str
                The element the atom belongs to
            parent : str
                The element of the atom's parent (if the atom is a hydrogen)

            Returns
            -------
            radius : str
                HCT radii
            scale : str
                Overlap factor
            """

            PARAMS_PATH = get_data_file_path('parameters/OBCparam.json')

            # Get rid of terminal white spaces
            atom_name = atom_name.strip()
            atom_type = atom_type.strip()

            # Load the dictioaries with the OBC parameters
            with open(PARAMS_PATH) as fd:
                params_by_type, scale_by_element, \
                    radius_by_element = json.load(fd)

            # Assign scale factor and HCT radius using the atom type
            if atom_type in params_by_type:
                scale, radius = params_by_type[atom_type]

            # Assign scale factor and HCT radius using the atom element
            else:
                found = True

                # Assign scale factor
                if element.upper() in scale_by_element:
                    scale = scale_by_element[element.upper()]
                    scale = _check_bonds(scale, element.upper(),
                                         degree, parent)
                else:
                    found = False

                # Assign scale factor
                if element.upper() in radius_by_element:
                    radius = radius_by_element[element.upper()]
                else:
                    found = False

                # Returns scale and HCT radius if found, otherwise it raises a
                # warning and returns the default parameters
                if not found:
                    log = Logger()
                    log.warning('Warning: OBC parameters for '
                                + '{} {} '.format(atom_name, atom_type)
                                + 'NOT found in the template '
                                + 'database. Using default parameters')
                    radius, scale = float(0.80), float(2.0)

            return radius, scale

        from simtk import unit

        # Loop over atom types and names:
        radii = list()
        scales = list()
        for atom_name, atom_type in zip(OPLS_params['atom_names'],
                                        OPLS_params['atom_types']):
            atom_name = re.sub('_', ' ', atom_name)
            radius, scale = _find_GBSA_parameters_according_to(
                atom_name, atom_type,
                degree_by_name.get(atom_name),
                element_by_name.get(atom_name),
                parent_by_name.get(atom_name))

            radii.append(unit.Quantity(float(radius), unit.angstrom),)
            scales.append(scale)

        # Assign OBC parameters
        OPLS_params['GBSA_radii'] = radii
        OPLS_params['GBSA_scales'] = scales


class OpenFFOPLS2005ParameterWrapper(BaseParameterWrapper):
    """
    It defines a parameters wrapper for an hybrid OpenFF-OPLS2005 force
    field.
    """

    _name = 'Openff + OPLS2005'



"""
This module contains miscellaneous set of handy classes and functions.
"""


__all__ = ["get_data_file_path",
           "temporary_cd",
           "warning_on_one_line",
           "check_if_path_exists",
           "create_path",
           "unit_to_string",
           "quantity_to_string",
           "convert_all_quantities_to_string",
           "string_to_quantity",
           "parse_charges_from_mae",
           "Logger"
           ]


from pkg_resources import resource_filename
import os
import contextlib
from simtk import unit


def get_data_file_path(relative_path):
    """
    It returns the path in the package's data location.

    Parameters
    ----------
    relative_path : str
        The relative path to the file that is required

    Returns
    -------
    output_path : str
        The path in the package's data location, if found
    """
    output_path = resource_filename('peleffy', os.path.join(
        'data', relative_path))

    if not os.path.exists(output_path):
        raise ValueError(
            "Sorry! {} does not exist. ".format(output_path)
            + "If you just added it, you'll have to re-install")

    return output_path


@contextlib.contextmanager
def temporary_cd(path):
    """
    A context that applies a temporary move to certain path.

    Parameters
    ----------
    path : str
        The path to move to
    """
    old_path = os.getcwd()
    os.chdir(os.path.abspath(path))
    try:
        yield
    finally:
        os.chdir(old_path)


def warning_on_one_line(message, category, filename, lineno, file=None,
                        line=None):
    """A customized warning output in a single line."""
    return ' %s:%s: %s:%s' % (filename, lineno, category.__name__, message)


def check_if_path_exists(path):
    """
    It checks if the supplied path exists and raises a ValueError otherwise.

    Parameters
    ----------
    path : str or pathlib.Path
        The path to check
    """
    from pathlib import Path

    if not isinstance(path, Path):
        path = Path(path)

    if not path.is_dir():
        raise ValueError('Invalid path to {}'.format(path))


def create_path(path):
    """
    It creates a path.

    Parameters
    ----------
    path : str
        The path that will be created
    """
    os.makedirs(str(path), exist_ok=True)


def unit_to_string(input_unit):
    """
    Serialize a simtk.unit.Unit and return it as a string.

    Function modified from the OpenFF Toolkit
    (https://github.com/openforcefield/openforcefield/).

    Parameters
    ----------
    input_unit : A simtk.unit
        The unit to serialize

    Returns
    -------
    unit_string : str
        The serialized unit.
    """

    if input_unit == unit.dimensionless:
        return "dimensionless"

    # Decompose output_unit into a tuples of (base_dimension_unit, exponent)
    unit_string = None
    for unit_component in input_unit.iter_base_or_scaled_units():
        unit_component_name = unit_component[0].name

        # Convert, for example "elementary charge" --> "elementary_charge"
        unit_component_name = unit_component_name.replace(' ', '_')
        if unit_component[1] == 1:
            contribution = '{}'.format(unit_component_name)
        else:
            contribution = '{}**{}'.format(unit_component_name,
                                           int(unit_component[1]))
        if unit_string is None:
            unit_string = contribution

        else:
            unit_string += ' * {}'.format(contribution)

    return unit_string


def quantity_to_string(input_quantity):
    """
    Serialize a simtk.unit.Quantity to a string.

    Function modified from the OpenFF Toolkit
    (https://github.com/openforcefield/openforcefield/).

    Parameters
    ----------
    input_quantity : simtk.unit.Quantity
        The quantity to serialize

    Returns
    -------
    output_string : str
        The serialized quantity
    """

    import numpy as np

    if input_quantity is None:
        return None

    unitless_value = input_quantity.value_in_unit(input_quantity.unit)

    # The string representation of a numpy array doesn't have commas and
    # breaks the parser, thus we convert any arrays to list here
    if isinstance(unitless_value, np.ndarray):
        unitless_value = list(unitless_value)

    unit_string = unit_to_string(input_quantity.unit)
    output_string = '{} * {}'.format(unitless_value, unit_string)

    return output_string


def convert_all_quantities_to_string(data_structure):
    """
    Traverses a data structure, attempting to convert all quantities
    into strings.

    Function modified from the OpenFF Toolkit
    (https://github.com/openforcefield/openforcefield/).

    Parameters
    ----------
    data_structure : dict
        A hierarchical dict structured

    Returns
    -------
    converted_data_structure : dict
        A hierarchical dict structured with simtk.unit.Quantitys
        converted to string
    """
    from copy import copy
    from peleffy.forcefield.parameters import BaseParameterWrapper

    data_structure = copy(data_structure)

    if isinstance(data_structure, BaseParameterWrapper):
        data_structure = dict(data_structure)

    if isinstance(data_structure, dict):
        for key, value in data_structure.items():
            data_structure[key] = convert_all_quantities_to_string(value)
        obj_to_return = data_structure

    elif isinstance(data_structure, list):
        for index, item in enumerate(data_structure):
            data_structure[index] = convert_all_quantities_to_string(item)
        obj_to_return = data_structure

    elif isinstance(data_structure, unit.Quantity):
        obj_to_return = quantity_to_string(data_structure)

    else:
        obj_to_return = data_structure

    return obj_to_return


def _ast_eval(node):
    """
    Performs an algebraic syntax tree evaluation of a unit.

    Function modified from the OpenFF Toolkit
    (https://github.com/openforcefield/openforcefield/).

    Parameters
    ----------
    node : An ast parsing tree node
    """
    import ast
    import operator as op

    operators = {
        ast.Add: op.add,
        ast.Sub: op.sub,
        ast.Mult: op.mul,
        ast.Div: op.truediv,
        ast.Pow: op.pow,
        ast.BitXor: op.xor,
        ast.USub: op.neg, }

    if isinstance(node, ast.Num):  # <number>
        return node.n
    elif isinstance(node, ast.BinOp):  # <left> <operator> <right>
        return operators[type(node.op)](_ast_eval(node.left),
                                        _ast_eval(node.right))
    elif isinstance(node, ast.UnaryOp):  # <operator> <operand> e.g., -1
        return operators[type(node.op)](_ast_eval(node.operand))
    elif isinstance(node, ast.Name):
        # see if this is a simtk unit
        b = getattr(unit, node.id)
        return b
    elif isinstance(node, ast.List):
        return ast.literal_eval(node)
    else:
        raise TypeError(node)


def string_to_quantity(quantity_string):
    """
    Takes a string representation of a quantity and returns a unit.Quantity

    Function modified from the OpenFF Toolkit
    (https://github.com/openforcefield/openforcefield/).

    Parameters
    ----------
    quantity_string : str
        The quantity to deserialize

    Returns
    -------
    output_quantity : simtk.unit.Quantity
        The deserialized quantity
    """
    import ast

    if quantity_string is None:
        return None

    output_quantity = _ast_eval(ast.parse(quantity_string, mode="eval").body)
    return output_quantity


def parse_charges_from_mae(path, parameters):
    """
    It reads an external file containing the partial charges to assign to the
    Molecule representation and updates the BaseParameterWrapper object
    containing the molecule's parameters.

    Parameters
    ----------
    path : str
        Path to the MAE file with the charges information.
    params : a BaseParameterWrapper object
        The BaseParameterWrapper object to update the charges

    Returns
    -------
    parameters : a BaseParameterWrapper object
        The BaseParameterWrapper object with the updated charges.
    """
    import re

    # Read external file containing the partial charges information
    params_info, params_list = ([] for i in range(2))
    copy = False
    with open(path, 'r') as file:
        for line in file.readlines():
            if bool(re.match(r' m_atom\[(.*?)\] {', line)):
                copy = True
                type_data = 'info'
            if ':::' in line:
                type_data = 'params'
            if '}' in line:
                copy = False
            if copy is True and type_data == 'info':
                params_info.append(line)
            if copy is True and type_data == 'params':
                params_list.append(line)
        params_info = [p.replace('\n', '').strip() for p in params_info[1:]]
        params_list = [l.replace('"', '').split() for l in params_list[1:-1]]

    # Get the index of the atom name and charge from the parameter's list
    idx_charges, idx_atom_name = (None for i in range(2))
    for idx, line in enumerate(params_info):
        if 's_m_pdb_atom_name' in line:
            idx_atom_name = idx
        if 'r_m_charge1' in line:
            idx_charges = idx
    if idx_charges is None or idx_atom_name is None:
        raise ValueError(
            " {} does not contain charges information. ".format(path))

    # Creates a charges by atom name dictionary
    d = {}
    for line in params_list:
        d[line[idx_atom_name]] = line[idx_charges]

    # Update the charges in BaseParameterWrapper object
    new_charges_parameters = []
    for atom_name in parameters['atom_names']:
        atom_name = atom_name.replace('_', '')
        if atom_name in d:
            new_charges_parameters.append(unit.Quantity(
                value=float(d.get(atom_name)),
                unit=unit.elementary_charge))
        else:
            raise ValueError(
                "Molecule atom name {} does not match with ".format(atom_name)
                + "any external file atom name's.")
    parameters['charges'] = new_charges_parameters
    return parameters


class Logger(object):
    """
    It contains all the required methods to handle logging messages.
    """
    import logging
    DEFAULT_LEVEL = logging.INFO

    def __init__(self):
        """It initializes a Logger object"""
        import logging

        # Get peleffy logger and set level only the first time
        if 'peleffy_log' not in logging.root.manager.loggerDict:
            self._logger = logging.getLogger('peleffy_log')
            self._logger.setLevel(self.DEFAULT_LEVEL)
        else:
            self._logger = logging.getLogger('peleffy_log')

        # If no handler is found add stream handler
        if not len(self._logger.handlers):
            ch = logging.StreamHandler()
            ch.setLevel(self.DEFAULT_LEVEL)
            formatter = logging.Formatter('%(message)s')
            ch.setFormatter(formatter)
            self._logger.addHandler(ch)

    def set_level(self, level):
        """
        It sets the logging level.

        Parameters
        ----------
        level : str
            The logging level to set. One of [DEBUG, INFO, WARNING, ERROR,
            CRITICAL]
        """
        import logging

        if level.upper() == 'DEBUG':
            logging_level = logging.DEBUG
        elif level.upper() == 'INFO':
            logging_level = logging.INFO
        elif level.upper() == 'WARNING':
            logging_level = logging.WARNING
        elif level.upper() == 'ERROR':
            logging_level = logging.ERROR
        elif level.upper() == 'CRITICAL':
            logging_level = logging.CRITICAL
        else:
            raise ValueError('Invalid level type')

        self._logger.setLevel(logging_level)
        for handler in self._logger.handlers:
            handler.setLevel(logging_level)

    def get_level(self):
        """
        It gets the logging level.

        Returns
        -------
        level : str
            The logging level to set. One of [DEBUG, INFO, WARNING, ERROR,
            CRITICAL]
        """
        import logging
        level = self._logger.level

        if level == logging.DEBUG:
            return 'DEBUG'
        if level == logging.INFO:
            return 'INFO'
        if level == logging.WARNING:
            return 'WARNING'
        if level == logging.ERROR:
            return 'ERROR'
        if level == logging.CRITICAL:
            return 'CRITICAL'

    def debug(self, *messages):
        """
        It pulls a debug message.

        Parameters
        ----------
        messages : list[str]
            The list of messages to print
        """
        if len(messages) > 1:
            self._logger.debug(' '.join(map(str, messages)))
        else:
            self._logger.debug(messages[0])

    def info(self, *messages):
        """
        It pulls an info message.

        Parameters
        ----------
        messages : list[str]
            The list of messages to print
        """
        if len(messages) > 1:
            self._logger.info(' '.join(map(str, messages)))
        else:
            self._logger.info(messages[0])

    def warning(self, *messages):
        """
        It pulls a warning message.

        Parameters
        ----------
        messages : list[str]
            The list of messages to print
        """
        if len(messages) > 1:
            self._logger.warning(' '.join(map(str, messages)))
        else:
            self._logger.warning(messages[0])

    def error(self, *messages):
        """
        It pulls a error message.

        Parameters
        ----------
        messages : list[str]
            The list of messages to print
        """
        if len(messages) > 1:
            self._logger.error(' '.join(map(str, messages)))
        else:
            self._logger.error(messages[0])

    def critical(self, *messages):
        """
        It pulls a critical message.

        Parameters
        ----------
        messages : list[str]
            The list of messages to print
        """
        if len(messages) > 1:
            self._logger.critical(' '.join(map(str, messages)))
        else:
            self._logger.critical(messages[0])



"""
This module defines the basic topology elements of a molecule.
"""


from simtk import unit


class _TopologyElement(object):
    """
    A wrapper for any topological element.
    """

    _name = None
    _writable_attrs = []

    class TopologyIterator(object):
        """
        An iterator for topological elements that iterates over their
        attributes in an ordered way.

        It is useful when writing topological elements to file.
        """

        def __init__(self, top_el):
            """
            It initiates a TopologyIterator object.

            Parameters
            ----------
            top_el : a TopologyElement object
                The topology element to iterate on.
            """
            self._index = int(0)
            self._top_el = top_el

        def __next__(self):
            """
            It returns the next item for the iteration.

            Returns
            -------
            attr_name : str
                The name of the attribute
            attr_value : float
                The value of the attribute
            """
            if self._index == len(self._top_el._writable_attrs):
                raise StopIteration

            attr_name = self._top_el._writable_attrs[self._index]
            attr_value = getattr(self._top_el, attr_name)
            self._index += 1

            return attr_name, attr_value

    @property
    def name(self):
        """
        The name that this topological element has.

        Returns
        -------
        name : str
            The name of the topological element
        """
        return self._name

    @property
    def n_writable_attrs(self):
        """
        The number of writable attributes this topological element has.

        Returns
        -------
        n_writable_attrs : int
            The number of writable attributes
        """
        return len(self._writable_attrs)

    def __iter__(self):
        """
        It returns an instance of the TopologyIterator.

        Returns
        -------
        iterator : a TopologyIterator
            The TopologyIterator object
        """
        return self.TopologyIterator(self)

    def __repr__(self):
        """
        It returns the representation string of this topological element.

        Returns
        -------
        repr_string : str
            The representation string
        """
        repr_string = '{}('.format(self._name)
        attrs = [attr for attr in self]
        for attr_name, value in attrs[:-1]:
            repr_string += '{}={}, '.format(attr_name, value)
        repr_string += '{}={})'.format(*attrs[-1])

        return repr_string

    def __str__(self):
        """
        It returns the readable representation string of this topological
        element.

        Returns
        -------
        str_string : str
            The readable representation string
        """
        return self.__repr__()


class Atom(_TopologyElement):
    """
    It represents the properties of an atom.
    """

    _name = 'Atom'
    _writable_attrs = ['index', 'PDB_name', 'OPLS_type']

    def __init__(self, index=-1, core=None, OPLS_type=None, PDB_name=None,
                 unknown=None, x=None, y=None, z=None, sigma=None,
                 epsilon=None, charge=None, born_radius=None, SASA_radius=None,
                 nonpolar_gamma=None, nonpolar_alpha=None, parent=None):
        """
        It initializes an Atom object.

        Parameters
        ----------
        index : int
            The index of the atom
        core : bool
            Whether atom is in the core or in a branch
        OPLS_type : str
            The OPLS type of the atom
        PDB_name : str
            The PDB name of the atom
        unknown : int
            The unknown value of the atom
        x : float
            The x coordinate of the atom
        y : float
            The y coordinate of the atom
        z : float
            The z coordinate of the atom
        sigma : simtk.unit.Quantity
            The sigma parameter of the atom
        epsilon : simtk.unit.Quantity
            The epsilon parameter of the atom
        charge : simtk.unit.Quantity
            The partial charge parameter of the atom
        born_radius : simtk.unit.Quantity
            The SGB born parameter radius of the atom
        SASA_radius : simtk.unit.Quantity
            The SASA radius parameter of the atom
        nonpolar_gamma : simtk.unit.Quantity
            The nonpolar gamma parameter of the atom
        nonpolar_alpha : simtk.unit.Quantity
            The nonpolar alpha parameter of the atom
        parent : peleffy.topology.Atom
            The parent of the atom
        """
        self._index = index
        self._core = core
        self._OPLS_type = OPLS_type
        self._PDB_name = PDB_name
        self._unknown = unknown
        self._x = x
        self._y = y
        self._z = z
        self._sigma = sigma
        self._epsilon = epsilon
        self._charge = charge
        self._born_radius = born_radius  # Rad. Non Polar SGB
        self._SASA_radius = SASA_radius  # Rad. Non Polar Type
        self._nonpolar_gamma = nonpolar_gamma  # SGB Non Polar gamma
        self._nonpolar_alpha = nonpolar_alpha  # SGB Non Polar type
        self._parent = parent

    def set_index(self, index):
        """
        It sets the index of the atom.

        Parameters
        ----------
        index : int
            The index of this Atom object
        """
        self._index = index

    def set_as_core(self):
        """It sets the atom as core"""
        self._core = True

    def set_as_branch(self):
        """It sets the atom as branch"""
        self._core = False

    def set_parent(self, parent):
        """
        It sets the parent of the atom.

        Parameters
        ----------
        parent : peleffy.topology.Atom
            The parent of the atom
        """
        self._parent = parent

    def set_coords(self, coords):
        """
        It sets the coordinates of the atom.

        Parameters
        ----------
        coords : list
            The coordinates array to set to this Atom object
        """
        assert len(coords) == 3, '3D array is expected'

        self._x, self._y, self._z = coords

    def set_OPLS_type(self, OPLS_type):
        """
        It sets the OPLS type of the atom.

        Parameters
        ----------
        OPLS_type : str
            The OPLS type to set to this Atom object
        """
        self._OPLS_type = OPLS_type

    def set_sigma(self, sigma):
        """
        It sets the sigma of the atom.

        Parameters
        ----------
        sigma : float
            The sigma to set to this Atom object
        """
        self._sigma = sigma

    def set_epsilon(self, epsilon):
        """
        It sets the epsilon of the atom.

        Parameters
        ----------
        epsilon : float
            The epsilon to set to this Atom object
        """
        self._epsilon = epsilon

    def set_charge(self, charge):
        """
        It sets the charge of the atom.

        Parameters
        ----------
        charge : float
            The charge to set to this Atom object
        """
        self._charge = charge

    def set_born_radius(self, born_radius):
        """
        It sets the Born radius of the atom.

        Parameters
        ----------
        born_radius : float
            The Born radius to set to this Atom object
        """
        self._born_radius = born_radius

    def set_SASA_radius(self, SASA_radius):
        """
        It sets the SASA radius of the atom.

        Parameters
        ----------
        SASA_radius : float
            The SASA radius to set to this Atom object
        """
        self._SASA_radius = SASA_radius

    def set_nonpolar_gamma(self, nonpolar_gamma):
        """
        It sets the nonpolar gamma of the atom.

        Parameters
        ----------
        nonpolar_gamma : float
            The nonpolar gamma to set to this Atom object
        """
        self._nonpolar_gamma = nonpolar_gamma

    def set_nonpolar_alpha(self, nonpolar_alpha):
        """
        It sets the nonpolar alpha of the atom.

        Parameters
        ----------
        nonpolar_alpha : float
            The nonpolar alpha to set to this Atom object
        """
        self._nonpolar_alpha = nonpolar_alpha

    @property
    def index(self):
        """
        Atom's index.

        Returns
        -------
        index : int
            The index of this Atom object
        """
        return self._index

    @property
    def core(self):
        """
        Atom's core position.

        Returns
        -------
        core : bool
            Whether this Atom object is in the core or in a branch
        """
        return self._core

    @property
    def OPLS_type(self):
        """
        Atom's OPLS type.

        .. todo ::

           * Consider removing any reference to OPLS, if possible
             Otherwise, use SMIRks to find the best match

        Returns
        -------
        OPLS_type : str
            The OLPS type of this Atom object
        """
        return self._OPLS_type

    @property
    def PDB_name(self):
        """
        Atom's PDB name.

        .. todo ::

           * Consider removing any reference to OPLS, if possible
             Otherwise, use SMIRks to find the best match

        Returns
        -------
        PDB_name : str
            The PDB name of this Atom object
        """
        return self._PDB_name

    @property
    def unknown(self):
        """
        Atom's unknown int.

        .. todo ::

           * Review the actual purpose of this attribute in PELE

        Returns
        -------
        unknown : int
            The unknown int of this Atom object
        """
        return self._unknown

    @property
    def x(self):
        """
        Atom's x coordinate.

        Returns
        -------
        x : float
            The x coordinate of this Atom object
        """
        return self._x

    @property
    def y(self):
        """
        Atom's y coordinate.

        Returns
        -------
        y : float
            The y coordinate of this Atom object
        """
        return self._y

    @property
    def z(self):
        """
        Atom's z coordinate.

        Returns
        -------
        z : float
            The z coordinate of this Atom object
        """
        return self._z

    @property
    def sigma(self):
        """
        Atom's sigma.

        Returns
        -------
        sigma : simtk.unit.Quantity
            The sigma parameter of this Atom object
        """
        return self._sigma

    @property
    def epsilon(self):
        """
        Atom's epsilon.

        Returns
        -------
        epsilon : simtk.unit.Quantity
            The epsilon parameter of this Atom object
        """
        return self._epsilon

    @property
    def charge(self):
        """
        Atom's charge.

        Returns
        -------
        charge : simtk.unit.Quantity
            The charge parameter of this Atom object
        """
        return self._charge

    @property
    def born_radius(self):
        """
        Atom's born radius.

        Returns
        -------
        born_radius : simtk.unit.Quantity
            The SGB Born radius parameter of this Atom object
        """
        return self._born_radius

    @property
    def SASA_radius(self):
        """
        Atom's SASA radius.

        Returns
        -------
        SASA_radius : simtk.unit.Quantity
            The SASA radius parameter of this Atom object
        """
        return self._SASA_radius

    @property
    def nonpolar_gamma(self):
        """
        Atom's nonpolar gamma.

        Returns
        -------
        nonpolar_gamma : simtk.unit.Quantity
            The nonpolar gamma parameter of this Atom object
        """
        return self._nonpolar_gamma

    @property
    def nonpolar_alpha(self):
        """
        Atom's nonpolar alpha.

        Returns
        -------
        nonpolar_alpha : simtk.unit.Quantity
            The nonpolar alpha parameter of this Atom object
        """
        return self._nonpolar_alpha

    @property
    def parent(self):
        """
        Atom's parent.

        Returns
        -------
        parent : simtk.unit.Quantity
            The nonpolar gamma parameter of this Atom object
        """
        return self._parent


class DummyAtom(Atom):
    """
    It represents a dummy atom.
    """

    def __init__(self, index=-1, PDB_name='DUMM', parent=None):
        """
        It initializes a DummyAtom object.

        Parameters
        ----------
        index : int
            The index of the atom
        PDB_name : str
            The PDB name of the atom
        parent : peleffy.topology.Atom
            The parent of the atom
        """
        if parent is None:
            parent = self
        super().__init__(index, False, None, PDB_name, None, None, None, None,
                         None, None, None, None, None, None, None, parent)


class Bond(_TopologyElement):
    """
    It represents a bond in the topology.
    """

    _name = 'Bond'
    _writable_attrs = ['atom1_idx', 'atom2_idx', 'spring_constant', 'eq_dist']

    def __init__(self, index=-1, atom1_idx=None, atom2_idx=None,
                 spring_constant=None, eq_dist=None):
        """
        It initiates a Bond object.

        Parameters
        ----------
        index : int
            The index of this Bond object
        atom1_idx : int
            The index of the first atom involved in this Bond
        atom2_idx : int
            The index of the second atom involved in this Bond
        spring_constant : simtk.unit.Quantity
            The spring constant of this Bond
        eq_dist : simtk.unit.Quantity
            The equilibrium distance of this Bond
        """
        self._index = index
        self._atom1_idx = atom1_idx
        self._atom2_idx = atom2_idx
        self._spring_constant = spring_constant
        self._eq_dist = eq_dist

    def set_atom1_idx(self, index):
        """
        It sets atom1's index.

        Parameters
        ----------
        index : int
            The index of the first atom involved in this Bond
        """
        self._atom1_idx = index

    def set_atom2_idx(self, index):
        """
        It sets atom2's index.

        Parameters
        ----------
        index : int
            The index of the second atom involved in this Bond
        """
        self._atom2_idx = index

    @property
    def index(self):
        """
        Bond's index.

        Returns
        -------
        index : int
            The index of this Bond object
        """
        return self._index

    @property
    def atom1_idx(self):
        """
        Bond's atom1 index.

        Returns
        -------
        atom1_idx : int
            The index of the first atom involved in this Bond object
        """
        return self._atom1_idx

    @property
    def atom2_idx(self):
        """
        Bond's atom2 index.

        Returns
        -------
        atom2_idx : int
            The index of the second atom involved in this Bond object
        """
        return self._atom2_idx

    @property
    def spring_constant(self):
        """
        Bond's spring constant.

        Returns
        -------
        spring_constant : simtk.unit.Quantity
            The spring constant of this Bond object
        """
        return self._spring_constant

    @property
    def eq_dist(self):
        """
        Bond's equilibrium distance.

        Returns
        -------
        eq_dist : simtk.unit.Quantity
            The equilibrium distance of this Bond object
        """
        return self._eq_dist


class Angle(_TopologyElement):
    """
    It represents an angle in the topology.
    """

    _name = 'Angle'
    _writable_attrs = ['atom1_idx', 'atom2_idx', 'atom3_idx',
                       'spring_constant', 'eq_angle']

    def __init__(self, index=-1, atom1_idx=None, atom2_idx=None,
                 atom3_idx=None, spring_constant=None, eq_angle=None):
        """
        It initiates an Angle object.

        Parameters
        ----------
        index : int
            The index of this Angle object
        atom1_idx : int
            The index of the first atom involved in this Angle
        atom2_idx : int
            The index of the second atom involved in this Angle
        atom3_idx : int
            The index of the third atom involved in this Angle
        spring_constant : simtk.unit.Quantity
            The spring constant of this Angle
        eq_angle : simtk.unit.Quantity
            The equilibrium angle of this Angle
        """
        self._index = index
        self._atom1_idx = atom1_idx
        self._atom2_idx = atom2_idx
        self._atom3_idx = atom3_idx
        self._spring_constant = spring_constant
        self._eq_angle = eq_angle

    def set_atom1_idx(self, index):
        """
        It sets atom1's index.

        Parameters
        ----------
        index : int
            The index of the first atom involved in this Angle
        """
        self._atom1_idx = index

    def set_atom2_idx(self, index):
        """
        It sets atom2's index.

        Parameters
        ----------
        index : int
            The index of the second atom involved in this Angle
        """
        self._atom2_idx = index

    def set_atom3_idx(self, index):
        """
        It sets atom3's index.

        Parameters
        ----------
        index : int
            The index of the third atom involved in this Angle
        """
        self._atom3_idx = index

    @property
    def index(self):
        """
        Angle's index.

        Returns
        -------
        index : int
            The index of this Angle object
        """
        return self._index

    @property
    def atom1_idx(self):
        """
        Angle's atom1 index.

        Returns
        -------
        atom1_idx : int
            The index of the first atom involved in this Angle object
        """
        return self._atom1_idx

    @property
    def atom2_idx(self):
        """
        Angle's atom2 index.

        Returns
        -------
        atom1_idx : int
            The index of the second atom involved in this Angle object
        """
        return self._atom2_idx

    @property
    def atom3_idx(self):
        """
        Angle's atom3 index.

        Returns
        -------
        atom1_idx : int
            The index of the third atom involved in this Angle object
        """
        return self._atom3_idx

    @property
    def spring_constant(self):
        """
        Angle's spring constant.

        Returns
        -------
        spring_constant : simtk.unit.Quantity
            The spring constant of this Angle object
        """
        return self._spring_constant

    @property
    def eq_angle(self):
        """
        Angle's equilibrium angle.

        Returns
        -------
        eq_angle : simtk.unit.Quantity
            The equilibrium angle of this Angle object
        """
        return self._eq_angle


class Dihedral(_TopologyElement):
    """
    It represents a dihedral in the topology.

    It can be a proper or an improper dihedral.
    """

    _name = 'Dihedral'

    def __init__(self, index=-1, atom1_idx=None, atom2_idx=None,
                 atom3_idx=None, atom4_idx=None, periodicity=None,
                 prefactor=None, constant=None, phase=None):
        """
        It initiates an Dihedral object.

        Parameters
        ----------
        index : int
            The index of this Dihedral object
        atom1_idx : int
            The index of the first atom involved in this Dihedral
        atom2_idx : int
            The index of the second atom involved in this Dihedral
        atom3_idx : int
            The index of the third atom involved in this Dihedral
        atom4_idx : int
            The index of the fourth atom involved in this Dihedral
        periodicity : int
            The periodicity of this Dihedral
        prefactor : int
            The prefactor of this Dihedral
        constant : simtk.unit.Quantity
            The constant of this Dihedral
        phase : simtk.unit.Quantity
            The phase constant of this Dihedral
        """
        self._index = index
        self._atom1_idx = atom1_idx
        self._atom2_idx = atom2_idx
        self._atom3_idx = atom3_idx
        self._atom4_idx = atom4_idx
        self._periodicity = periodicity
        self._prefactor = prefactor
        self._constant = constant
        self._phase = phase

    def set_atom1_idx(self, index):
        """
        It sets atom1's index.

        Parameters
        ----------
        index : int
            The index of the first atom involved in this Dihedral
        """
        self._atom1_idx = index

    def set_atom2_idx(self, index):
        """
        It sets atom2's index.

        Parameters
        ----------
        index : int
            The index of the second atom involved in this Dihedral
        """
        self._atom2_idx = index

    def set_atom3_idx(self, index):
        """
        It sets atom3's index.

        Parameters
        ----------
        index : int
            The index of the third atom involved in this Dihedral
        """
        self._atom3_idx = index

    def set_atom4_idx(self, index):
        """
        It sets atom4's index.

        Parameters
        ----------
        index : int
            The index of the fourth atom involved in this Dihedral
        """
        self._atom4_idx = index

    def plot(self):
        """
        It plots this Dihedral as a function of phi angle.
        """
        from matplotlib import pyplot
        import numpy as np

        x = unit.Quantity(np.arange(0, np.pi, 0.1), unit=unit.radians)
        pyplot.plot(x, self.constant * (1 + self.prefactor
                                        * np.cos(self.periodicity * x
                                                 + self.phase)),
                    'r--')

        pyplot.show()

    @property
    def index(self):
        """
        Dihedral's index.

        Returns
        -------
        index : int
            The index of this Dihedral object
        """
        return self._index

    @property
    def atom1_idx(self):
        """
        Dihedral's atom1 index.

        Returns
        -------
        atom1_idx : int
            The index of the first atom involved in this Dihedral object
        """
        return self._atom1_idx

    @property
    def atom2_idx(self):
        """
        Dihedral's atom2 index.

        Returns
        -------
        atom1_idx : int
            The index of the second atom involved in this Dihedral object
        """
        return self._atom2_idx

    @property
    def atom3_idx(self):
        """
        Dihedral's atom3 index.

        Returns
        -------
        atom1_idx : int
            The index of the third atom involved in this Dihedral object
        """
        return self._atom3_idx

    @property
    def atom4_idx(self):
        """
        Dihedral's atom4 index.

        Returns
        -------
        atom1_idx : int
            The index of the fourth atom involved in this Dihedral object
        """
        return self._atom4_idx

    @property
    def periodicity(self):
        """
        Dihedral's periodicity.

        Returns
        -------
        periodicity : int
            The periodicity this Dihedral object
        """
        return self._periodicity

    @property
    def prefactor(self):
        """
        Dihedral's prefactor.

        Returns
        -------
        prefactor : int
            The prefactor this Dihedral object
        """
        return self._prefactor

    @property
    def constant(self):
        """
        Dihedral's constant.

        Returns
        -------
        constant : unit.simtk.Quantity
            The constant of this Dihedral object
        """
        return self._constant

    @property
    def phase(self):
        """
        Dihedral's phase constant.

        Returns
        -------
        phase : unit.simtk.Quantity
            The phase constant of this Dihedral object
        """
        return self._phase


class Proper(Dihedral):
    """
    It represents a proper dihedral in the topology.
    """

    _name = 'Proper'
    exclude = False
    _writable_attrs = ['atom1_idx', 'atom2_idx', 'atom3_idx', 'atom4_idx',
                       'constant', 'prefactor', 'periodicity', 'phase']

    def include_in_14_list(self):
        """
        It includes this proper dihedral in PELE's 1-4 list.
        """
        self.exclude = False

    def exclude_from_14_list(self):
        """
        It excludes this proper dihedral from PELE's 1-4 list by
        setting the index of the third atom to negative.
        """
        self.exclude = True


class Improper(Dihedral):
    """
    It represents an improper dihedral in the topology.
    """

    _name = 'Improper'
    _writable_attrs = ['atom1_idx', 'atom2_idx', 'atom3_idx', 'atom4_idx',
                       'constant', 'prefactor', 'periodicity']


class OFFDihedral(_TopologyElement):
    """
    It represents a dihedral in the Open Force Field's topology.
    """

    _name = 'OFFDihedral'
    _writable_attrs = ['atom1_idx', 'atom2_idx', 'atom3_idx', 'atom4_idx',
                       'periodicity', 'phase', 'k', 'idivf']
    _to_PELE_class = Dihedral

    def __init__(self, index=-1, atom1_idx=None, atom2_idx=None,
                 atom3_idx=None, atom4_idx=None, periodicity=None,
                 phase=None, k=None, idivf=None):
        """
        It initiates an Dihedral object.

        Parameters
        ----------
        index : int
            The index of this Dihedral object
        atom1_idx : int
            The index of the first atom involved in this Dihedral
        atom2_idx : int
            The index of the second atom involved in this Dihedral
        atom3_idx : int
            The index of the third atom involved in this Dihedral
        atom4_idx : int
            The index of the fourth atom involved in this Dihedral
        periodicity : int
            The periodicity of this Dihedral
        phase : simtk.unit.Quantity
            The phase of this Dihedral
        k : simtk.unit.Quantity
            The constant of this Dihedral
        idivf : int
            The idivf of this Dihedral
        """
        self.index = index
        self.atom1_idx = atom1_idx
        self.atom2_idx = atom2_idx
        self.atom3_idx = atom3_idx
        self.atom4_idx = atom4_idx
        self.periodicity = periodicity
        self.phase = phase
        self.k = k
        self.idivf = idivf

    def _check_up(self):
        """
        It performs some parameter check ups.

        Raises
        ------
        AssertionError
            If any unexpected value is found
        """
        pass

    def to_PELE(self):
        """
        It converts this Open Force Field Dihedral object into a
        PELE-compatible one.

        .. todo ::

           * Fully cover all OpenFF dihedral parameters

        Returns
        -------
        PELE_dihedral : a Dihedral
            The PELE-compatible Dihedral object
        """
        if (self.periodicity is None or self.phase is None
                or self.k is None or self.idivf is None):
            return None

        try:
            self._check_up()
        except AssertionError as e:
            raise ValueError('Invalid value found: {}'.format(e))

        PELE_phase = self.phase

        if self.phase.value_in_unit(unit.degree) == 180:
            PELE_prefactor = -1
            PELE_phase = unit.Quantity(value=0.0, unit=unit.degree)
        else:
            PELE_prefactor = 1

        PELE_constant = self.k / self.idivf

        PELE_dihedral_kwargs = {'index': self.index,
                                'atom1_idx': self.atom1_idx,
                                'atom2_idx': self.atom2_idx,
                                'atom3_idx': self.atom3_idx,
                                'atom4_idx': self.atom4_idx,
                                'periodicity': self.periodicity,
                                'prefactor': PELE_prefactor,
                                'constant': PELE_constant,
                                'phase': PELE_phase}

        return self._to_PELE_class(**PELE_dihedral_kwargs)

    def plot(self):
        """
        It plots this Dihedral as a function of phi angle.
        """
        from matplotlib import pyplot
        import numpy as np

        x = unit.Quantity(np.arange(0, np.pi, 0.1), unit=unit.radians)
        pyplot.plot(x,
                    self.k * (1 + np.cos(self.periodicity * x - self.phase)),
                    'r--')

        pyplot.show()


class OFFProper(OFFDihedral):
    """
    It represents a proper dihedral in the Open Force Field's topology.
    """

    _name = 'OFFProper'
    _to_PELE_class = Proper

    def _check_up(self):
        """
        It performs some parameter check ups.

        .. todo ::

            * Periodicity can also equal 5 and currently it is not supported
             by PELE

        Raises
        ------
        AssertionError
            If any unexpected value is found
        """
        assert self.periodicity in (1, 2, 3, 4, 5, 6), 'Expected values ' \
            'for periodicity are 1, 2, 3, 4, 5 or 6, obtained ' \
            '{}'.format(self.periodicity)

        # proper's idivfs must always be 1 --> Apparently not: COC(O)OC, t84, t86
        # assert self.idivf == 1, 'The expected value for idivf is 1 ' \
        #     'for propers, obtained {}'.format(self.idivf)

        # The next version of PELE will be compatible with phase values
        # other than 0 and 180 degrees to fully cover all OpenFF dihedrals
        # assert self.phase.value_in_unit(unit.degree) in (0, 180), \
        #     'Expected values for phase are 0 or 180, obtained ' \
        #     '{}'.format(self.phase)


class OFFImproper(OFFDihedral):
    """
    It represents an improper dihedral in the Open Force Field's topology.
    """

    _name = 'OFFImproper'
    _to_PELE_class = Improper

    def _check_up(self):
        """
        It performs some parameter check ups.

        .. todo ::

            * Periodicity can also equal 5 and currently it is not supported
             by PELE

        Raises
        ------
        AssertionError
            If any unexpected value is found
        """
        assert self.periodicity in (1, 2, 3, 4, 5, 6), 'Expected values ' \
            'for periodicity are 1, 2, 3, 4, 5 or 6, obtained ' \
            '{}'.format(self.periodicity)

        assert self.phase.value_in_unit(unit.degree) in (0, 180), \
            'Expected values for phase are 0 or 180 in impropers, ' \
            'obtained {}'.format(self.phase)



"""
This module contains integrity tests that check that all peleffy's modules
are available.
"""

import pytest


class TestIntegrity(object):
    """Integrity test."""

    def test_modules(self):
        try:
            import peleffy
            from peleffy import forcefield
            from peleffy import solvent
            from peleffy import template
            from peleffy import topology
            from peleffy import utils

        except ImportError as e:
            raise AssertionError('The following peleffy module is missing: '
                                 + str(e))



"""
This module contains different parameter calculators for force fields.
"""


from simtk import unit

from peleffy.utils.toolkits import ToolkitUnavailableException


class _PartialChargeCalculator(object):
    """
    Base class for partial charge calculators.
    """

    from peleffy.utils.toolkits import AmberToolkitWrapper

    _name = None
    _amber_toolkit = AmberToolkitWrapper()

    def __init__(self, molecule):
        """
        It initiates a PartialChargeCalculator object.

        Parameters
        ----------
        molecule : a peleffy.topology.Molecule
            The partial charges of this Molecule object will be calculated
        """
        self._molecule = molecule

    @property
    def molecule(self):
        """
        The peleffy's Molecule.

        Returns
        -------
        molecule : a peleffy.topology.Molecule
            The peleffy's Molecule object
        """
        return self._molecule

    @property
    def name(self):
        return self._name

    def get_partial_charges(self):
        """
        It returns the partial charges that correspond to the molecule's
        atoms.

        Returns
        -------
        charges : simtk.unit.Quantity
            The array of partial charges
        """

        return self._amber_toolkit.compute_partial_charges(self.molecule,
                                                           method=self.name)

    def assign_partial_charges(self, parameters):
        """
        Given a parameter wrapper, it computes the partial charges and
        assigns them to it.

        Parameters
        ----------
        parameters : a peleffy.forcefield.parameters.BaseParameterWrapper object
            The parameter wrapper containing the parameters generated
            with the current force field and where the partial charges
            will be assigned
        """

        partial_charges = self.get_partial_charges()
        parameters['charges'] = partial_charges


class Am1bccCalculator(_PartialChargeCalculator):
    """
    Implementation of the AM1-BCC partial charge calculator (using RDKit).
    """

    _name = 'am1bcc'


class GasteigerCalculator(_PartialChargeCalculator):
    """
    Implementation of the gasteiger partial charge calculator (using
    RDKit).
    """

    _name = 'gasteiger'


class OPLSChargeCalculator(_PartialChargeCalculator):
    """
    Implementation of the calculator of OPLS partial charges (using
    Schrodinger's ffld_server)
    """

    _name = 'OPLS2005'

    def get_partial_charges(self):
        """
        It returns the partial charges that correspond to the molecule's
        atoms.

        Returns
        -------
        partial_charges : simtk.unit.Quantity
            The array of partial charges
        """

        from peleffy.utils.toolkits import SchrodingerToolkitWrapper

        try:
            schrodinger_toolkit_wrapper = SchrodingerToolkitWrapper()
            ffld_output = schrodinger_toolkit_wrapper.run_ffld_server(
                self.molecule)
        except ToolkitUnavailableException:
            raise ToolkitUnavailableException(
                'OPLSChargeCalculator requires the Schrodinger '
                + 'Toolkit to obtain partial charges')

        from peleffy.forcefield.parameters import OPLS2005ParameterWrapper
        parameters = OPLS2005ParameterWrapper.from_ffld_output(
            self.molecule, ffld_output)

        return parameters['charges']

    def assign_partial_charges(self, parameters):
        """
        Given a parameter wrapper, it computes the partial charges and
        assigns them to it.

        Parameters
        ----------
        parameters : a peleffy.forcefield.parameters.BaseParameterWrapper object
            The parameter wrapper containing the parameters generated
            with the current force field and where the partial charges
            will be assigned
        """
        import peleffy

        if not isinstance(
                parameters,
                peleffy.forcefield.parameters.OPLS2005ParameterWrapper):
            partial_charges = self.get_partial_charges()
            parameters['charges'] = partial_charges


class DummyChargeCalculator(_PartialChargeCalculator):
    """
    Implementation of a dummy charge calculator that will not perform
    any partial charge calculation.
    """

    _name = 'OPLS2005'

    def get_partial_charges(self):
        """
        It returns the partial charges that correspond to the molecule's
        atoms.

        Returns
        -------
        partial_charges : simtk.unit.Quantity
            The array of partial charges
        """

        # Extract number of atoms in molecule
        n_atoms = len(self.molecule.get_pdb_atom_names())

        return [unit.Quantity(0.0, unit.elementary_charge), ] * n_atoms

    def assign_partial_charges(self, parameters):
        """
        Given a parameter wrapper, it computes the partial charges and
        assigns them to it.

        Parameters
        ----------
        parameters : a peleffy.forcefield.parameters.BaseParameterWrapper object
            The parameter wrapper containing the parameters generated
            with the current force field and where the partial charges
            will be assigned
        """
        import peleffy

        if not isinstance(
                parameters,
                peleffy.forcefield.parameters.OPLS2005ParameterWrapper):
            partial_charges = self.get_partial_charges()
            parameters['charges'] = partial_charges



import os
from os.path import relpath, join
from setuptools import setup
import versioneer


def find_package_data(data_root, package_root):
    files = []
    for root, dirnames, filenames in os.walk(data_root):
        for fn in filenames:
            files.append(relpath(join(root, fn), package_root))
    return files


setup(
    name="peleffy",
    author="Martí Municoy",
    author_email="martimunicoy@gmail.com",
    description=("PELE Force Field Yielder"),
    license="MIT",
    keywords="molecular mechanics, forcefield, potential energy",
    url="http://github.com/martimunicoy/peleffy",
    packages=[
        'peleffy',
        'peleffy/tests',
        'peleffy/data',
        'peleffy/template',
        'peleffy/solvent',
        'peleffy/topology',
        'peleffy/forcefield',
        'peleffy/utils'
    ],
    long_description='The peleffy (PELE Force Field Yielder) is '
    + 'a Python package that builds PELE-compatible force '
    + 'field templates.',
    classifiers=[
        "Development Status :: 1 - Planning",
        "Natural Language :: English",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix"
    ],
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    package_data={'peleffy': find_package_data(
        'peleffy/data', 'peleffy')},
)
